"""Pipeline to gather new data to be imported into a MySQL database."""

import argparse
import csv
from datetime import datetime, date
import os
import pathlib
import sys

from Bio import SeqIO

from pdm_utils.classes.alchemyhandler import AlchemyHandler
from pdm_utils.classes import genomepair
from pdm_utils.classes import ticket
from pdm_utils.constants import constants
from pdm_utils.functions import basic
from pdm_utils.functions import configfile
from pdm_utils.functions import fileio
from pdm_utils.functions import ncbi
from pdm_utils.functions import mysqldb
from pdm_utils.functions import phagesdb
from pdm_utils.functions import tickets

# Names of folders and files created.
DEFAULT_OUTPUT_FOLDER = os.getcwd()
CURRENT_DATE = date.today().strftime("%Y%m%d")
RESULTS_FOLDER = f"{CURRENT_DATE}_get_data"
PECAAN_FOLDER = "pecaan"
GENBANK_FOLDER = "genbank"
PHAGESDB_FOLDER = "phagesdb"
UPDATES_FOLDER = "updates"
UPDATE_TABLE = "update_table.csv"
IMPORT_TABLE = "import_table.csv"
GENBANK_RESULTS_TABLE = "genbank_results.csv"
GENOME_FOLDER = "genomes"

# Columns for import ticket table.
IMPORT_COLUMNS = constants.IMPORT_TABLE_STRUCTURE["order"]

# Columns for update ticket table.
UPDATE_COLUMNS = ["table",
                  "field",
                  "value",
                  "key_name",
                  "key_value"]

# Column headers for GenBank results file.
NCBI_RESULTS_COLUMNS = ["phage_id",
                       "phage_name",
                       "accession",
                       "annotation_status",
                       "mysql_date",
                       "genbank_date",
                       "result"]

# GenBank retrieval status options
NOT_AUTO = "not_auto_updated"
NO_ACC = "no_accession"
DUPE_ACC = "duplicate_accession"
OLD_DOCSUM = "docsum_not_new"
OLD_RECORD = "record_not_new"
RET_FAIL = "retrieval_failure"
RETRIEVED = "retrieved_for_import"
GENBANK_STATUS = {NOT_AUTO, NO_ACC, DUPE_ACC, OLD_DOCSUM,
                  OLD_RECORD, RET_FAIL, RETRIEVED}

# TODO unittest.
def parse_args(unparsed_args_list):
    """Verify the correct arguments are selected for getting updates."""

    retrieve_help = "Pipeline to get new data to import into a database."
    database_help = "Name of the MySQL database."
    output_folder_help = "Path to the directory where updates will be stored."
    updates_help = "Retrieve field updates from PhagesDB."
    draft_help = "Retrieve auto-annotated 'draft' genomes from PECAAN."
    final_help = "Retrieve manually-annotated 'final' genomes from PhagesDB."
    genbank_help = "Retrieve annotated genomes from GenBank."
    all_help = "Retrieve all types of new data."
    genbank_results_help = "Store results of Genbank record retrieval."
    force_download_help = "Retrieve genomes regardless of date in database."
    config_file_help = "Path to the file containing user-specific login data."

    parser = argparse.ArgumentParser(description=retrieve_help)
    parser.add_argument("database", type=str, help=database_help)
    parser.add_argument("-o", "--output_folder", type=pathlib.Path,
        default=pathlib.Path(DEFAULT_OUTPUT_FOLDER), help=output_folder_help)
    parser.add_argument("-u", "--updates", action="store_true",
        default=False, help=updates_help)
    parser.add_argument("-d", "--draft", action="store_true",
        default=False, help=draft_help)
    parser.add_argument("-f", "--final", action="store_true",
        default=False, help=final_help)
    parser.add_argument("-g", "--genbank", action="store_true",
        default=False, help=genbank_help)
    parser.add_argument("-a", "--all_data", action="store_true",
        default=False, help=all_help)
    parser.add_argument("-gr", "--genbank_results", action="store_true",
        default=False, help=genbank_results_help)
    parser.add_argument("-fd", "--force_download", action="store_true",
        default=False, help=force_download_help)
    parser.add_argument("-c", "--config_file", type=pathlib.Path,
                        help=config_file_help, default=None)

    # Assumed command line arg structure:
    # python3 -m pdm_utils.run <pipeline> <additional args...>
    # sys.argv:      [0]            [1]         [2...]
    args = parser.parse_args(unparsed_args_list[2:])

    if (args.updates == False and
            args.draft == False and
            args.final == False and
            args.genbank == False):
        args.all_data = True

    if args.all_data == True:
        args.updates = True
        args.draft = True
        args.final = True
        args.genbank = True

    if args.genbank == False:
        args.genbank_results = False

    return args

# TODO unittest.
def main(unparsed_args_list):
    """Run main retrieve_updates pipeline."""
    # Parse command line arguments
    args = parse_args(unparsed_args_list)
    force = args.force_download
    args.output_folder = basic.set_path(args.output_folder, kind="dir",
                                        expect=True)
    working_dir = pathlib.Path(RESULTS_FOLDER)
    working_path = basic.make_new_dir(args.output_folder, working_dir,
                                      attempt=50)

    if working_path is None:
        print(f"Invalid working directory '{working_dir}'")
        sys.exit(1)

    # Create config object with data obtained from file and/or defaults.
    config = configfile.build_complete_config(args.config_file)
    mysql_creds = config["mysql"]
    ncbi_creds = config["ncbi"]

    # Verify database connection and schema compatibility.
    print("Preparing genome data sets from the MySQL database...")
    alchemist = AlchemyHandler(database=args.database,
                               username=mysql_creds["user"],
                               password=mysql_creds["password"])
    alchemist.connect(pipeline=True)
    engine = alchemist.engine
    mysqldb.check_schema_compatibility(engine, "the get_data pipeline")

    # Get existing data from MySQL to determine what needs to be updated.
    query = ("SELECT PhageID, Name, HostGenus, Status, Cluster, "
             "DateLastModified, Accession, RetrieveRecord, Subcluster, "
             "AnnotationAuthor FROM phage")

    mysqldb_genome_list =  mysqldb.parse_genome_data(engine=engine,
                                                     phage_query=query,
                                                     gnm_type="mysqldb")
    engine.dispose()
    mysqldb_genome_dict = {}
    for gnm in mysqldb_genome_list:
        # With default date, the date of all records retrieved will be newer.
        if force:
            gnm.date = constants.EMPTY_DATE
        mysqldb_genome_dict[gnm.id] = gnm

    # Get data from PhagesDB
    if (args.updates or args.final or args.draft) is True:
        print("Retrieving data from PhagesDB...")
        phagesdb_phages = phagesdb.get_phagesdb_data(constants.API_SEQUENCED)
        phagesdb_phages_dict = basic.convert_list_to_dict(phagesdb_phages,
                                                          "phage_name")
        phagesdb_genome_dict = phagesdb.parse_genomes_dict(
                                    phagesdb_phages_dict,
                                    gnm_type="phagesdb",
                                    seq=False)

        # Exit if all phage data wasn't retrieved.
        if len(phagesdb_genome_dict) == 0:
            sys.exit(1)

        # Returns a list of tuples.
        tup = match_genomes(mysqldb_genome_dict, phagesdb_genome_dict)
        matched_genomes = tup[0]
        unmatched_phagesdb_ids = tup[1]

    if args.updates is True:
        get_update_data(working_path, matched_genomes)
    if args.final is True:
        get_final_data(working_path, matched_genomes)
    if args.genbank is True:
        get_genbank_data(working_path, mysqldb_genome_dict,
                         ncbi_creds, args.genbank_results, force=force)
    if args.draft is True:
        if force:
            # Add all draft genomes currently in database to the list of
            # draft genomes to be downloaded.
            drafts = get_matched_drafts(matched_genomes)
            unmatched_phagesdb_ids |= drafts
        get_draft_data(working_path, unmatched_phagesdb_ids)


# TODO unittest.
def get_matched_drafts(matched_genomes):
    """Generate a list of matched 'draft' genomes."""
    # matched data = list of gnm_pair objects
    # unmatched data = set of PhageIDs
    drafts = set()
    for gnm_pair in matched_genomes:
        mysqldb_gnm = gnm_pair.genome1
        phagesdb_gnm = gnm_pair.genome2
        if mysqldb_gnm.annotation_status == "draft":
            # Since draft phage ids are used by PECAAN to retrieve data from
            # PhagesDB, use the id from the PhagesDB genome object.
            drafts.add(phagesdb_gnm.id)
    return drafts

# TODO unittest.
def match_genomes(dict1, dict2):
    """Match MySQL database genome data to PhagesDB genome data.

    Both dictionaries:
    Key = PhageID
    Value = pdm_utils genome object"""

    # Generate phage_id sets and match sets.
    d2_keys = dict2.keys()
    d1_keys = dict1.keys()
    matched_keys = d1_keys & d2_keys
    d1_unmatched_keys = d1_keys - d2_keys
    d2_unmatched_keys = d2_keys - d1_keys

    matched_genomes = []
    for key in matched_keys:
        gnm_pair = genomepair.GenomePair()
        gnm_pair.genome1 = dict1[key]
        gnm_pair.genome2 = dict2[key]
        matched_genomes.append(gnm_pair)

    # Only unmatched with AnnotationAuthor = 1
    unmatched_d1_genomes = {}
    for key in d1_unmatched_keys:
        gnm = dict1[key]
        if gnm.annotation_author == 1:
            unmatched_d1_genomes[key] = gnm

    results = {"match_count": len(matched_keys),
               "d1_unmatch_count": len(d1_unmatched_keys),
               "d2_unmatch_count": len(d2_unmatched_keys),
               "d1_unmatch_aa1": unmatched_d1_genomes.keys()
               }
    print_match_results(results)

    return (matched_genomes, d2_unmatched_keys)


# TODO unittest
def print_match_results(dict):
    """Print results of genome matching."""
    print("\nSummary of genome matching:")
    print(f"{dict['match_count']:>6}: genome(s) matched.")
    print(f"{dict['d1_unmatch_count']:>6}: MySQL genome(s) not matched.")
    print(f"{dict['d2_unmatch_count']:>6}: PhagesDB genome(s) not matched.")
    count = len(dict["d1_unmatch_aa1"])
    if count > 0:
        print(f"{count} MySQL genome(s) that are unexpectedly unmatched:")
        for name in dict["d1_unmatch_aa1"]:
            print(name)


# TODO unittest
def get_update_data(output_folder, matched_genomes):
    """Run sub-pipeline to retrieve field updates from PhagesDB."""
    updates_folder = pathlib.Path(output_folder, UPDATES_FOLDER)
    updates_folder.mkdir()
    update_tickets = []
    for gnm_pair in matched_genomes:
        tkt_list = compare_data(gnm_pair)
        update_tickets.extend(tkt_list)

    # Field updates
    if len(update_tickets) > 0:
        print(f"\n\n{len(update_tickets)} field updates are available.")
        filepath = pathlib.Path(updates_folder, UPDATE_TABLE)
        fileio.export_data_dict(update_tickets, filepath, UPDATE_COLUMNS,
                               include_headers=True)
    else:
        print("\n\nNo field updates.")

    # Now remove empty folders.
    if len(basic.identify_contents(updates_folder, kind=None)) == 0:
        updates_folder.rmdir()



# TODO test.
def compare_data(gnm_pair):
    """Compare data and create update tickets."""
    tkt_list = []
    mysqldb_gnm = gnm_pair.genome1
    phagesdb_gnm = gnm_pair.genome2

    # Compare Cluster
    if mysqldb_gnm.cluster != phagesdb_gnm.cluster:
        # "Singleton" is not a valid cluster in MySQL (should be NULL)
        if phagesdb_gnm.cluster == "Singleton":
            # Convert to "NULL" so pipelines.update_field.update_field()
            # can convert to None
            phagesdb_gnm.cluster = "NULL"
        d1 = create_update_ticket("Cluster", phagesdb_gnm.cluster,
                                  mysqldb_gnm.id)
        tkt_list.append(d1)

    # Compare Subcluster
    if mysqldb_gnm.subcluster != phagesdb_gnm.subcluster:
        d2 = create_update_ticket("Subcluster", phagesdb_gnm.subcluster,
                                  mysqldb_gnm.id)
        tkt_list.append(d2)

    # Compare Host genus
    if mysqldb_gnm.host_genus != phagesdb_gnm.host_genus:
        d3 = create_update_ticket("HostGenus", phagesdb_gnm.host_genus,
                                  mysqldb_gnm.id)
        tkt_list.append(d3)

    # Compare Accession
    # If the genome author is not 1 ("hatfull"), then don't worry about
    # updating the accession. This used to be determined with
    # the status field, but now it is determined with the
    # AnnotationAuthor field.
    if (mysqldb_gnm.accession != phagesdb_gnm.accession and \
            mysqldb_gnm.annotation_author == 1):
        d4 = create_update_ticket("Accession", phagesdb_gnm.accession,
                                  mysqldb_gnm.id)
        tkt_list.append(d4)

    return tkt_list

# TODO test.
def create_update_ticket(field, value, key_value):
    """Create update ticket."""
    dict = {"table":"phage",
            "field":field,
            "value":value,
            "key_name":"PhageID",
            "key_value":key_value}
    return dict


# TODO unittest
def convert_tickets_to_dict(list_of_tickets):
    """Convert list of tickets to list of dictionaries."""
    l = []
    for tkt in list_of_tickets:
        dict = {
            "type": tkt.type,
            "phage_id": tkt.phage_id,
            "host_genus": tkt.data_dict["host_genus"],
            "cluster": tkt.data_dict["cluster"],
            "subcluster": tkt.data_dict["subcluster"],
            "annotation_status": tkt.data_dict["annotation_status"],
            "annotation_author": tkt.data_dict["annotation_author"],
            "description_field": tkt.description_field,
            "accession": tkt.data_dict["accession"],
            "eval_mode": tkt.eval_mode,
            "retrieve_record": tkt.data_dict["retrieve_record"]
            }
        l.append(dict)
    return l


# TODO unittest
def get_final_data(output_folder, matched_genomes):
    """Run sub-pipeline to retrieve 'final' genomes from PhagesDB."""

    print(f"\n\nDownloading genome(s) from PhagesDB.")
    phagesdb_folder = pathlib.Path(output_folder, PHAGESDB_FOLDER)
    phagesdb_folder.mkdir()
    genome_folder = pathlib.Path(phagesdb_folder, GENOME_FOLDER)
    genome_folder.mkdir()
    import_tickets = []
    failed_list = []

    # Iterate through each phage in the MySQL database
    for gnm_pair in matched_genomes:
        mysqldb_gnm = gnm_pair.genome1
        phagesdb_gnm = gnm_pair.genome2

        # Not all phages have associated Genbank-formatted files
        # available on PhagesDB. Check to see if there is a flatfile for
        # this phage. Download the flatfile only if there is a date tag,
        # and only if that date is more recent than the date stored in
        # the MySQL database for that genome. The tagged date only reflects when
        # the file was uploaded into PhagesDB. The date the actual
        # Genbank record was created is stored within the file,
        # and this too could be less recent than the current version in
        # the MySQL database; however, this part gets checked during the import
        # stage.
        set_phagesdb_gnm_date(phagesdb_gnm)
        set_phagesdb_gnm_file(phagesdb_gnm)
        if (phagesdb_gnm.filename != "" and phagesdb_gnm.date > mysqldb_gnm.date):
            # Save the file on the hard drive with the same name as
            # stored on PhagesDB
            flatfile_data = phagesdb.retrieve_url_data(phagesdb_gnm.filename)
            if flatfile_data == "":
                failed_list.append(mysqldb_gnm.id)
            else:
                save_phagesdb_file(flatfile_data, phagesdb_gnm, genome_folder)
                tkt = create_phagesdb_ticket(mysqldb_gnm.id)
                import_tickets.append(tkt)

    if len(import_tickets) > 0:
        print(f"\n\n{len(import_tickets)} genome(s) "
              "were retrieved from PhagesDB.")
        create_ticket_table(import_tickets, phagesdb_folder)

    if len(failed_list) > 0:
        print(f"{len(failed_list)} genome(s) failed to be retrieved:")
        for element in failed_list:
            print(element)
        input("\n\nPress ENTER to continue.")

    # Now remove empty folders.
    if len(basic.identify_contents(genome_folder, kind=None)) == 0:
        genome_folder.rmdir()
    if len(basic.identify_contents(phagesdb_folder, kind=None)) == 0:
        phagesdb_folder.rmdir()


# TODO unittest.
def save_phagesdb_file(data, gnm, output_folder):
    """Save file retrieved from PhagesDB."""
    filename = gnm.filename.split("/")[-1]
    filepath = pathlib.Path(output_folder, filename)
    with filepath.open("w") as fh:
        fh.write(data)

# TODO test.
def save_pecaan_file(response, name, output_folder):
    """Save data retrieved from PECAAN."""
    filename = f"{name}.txt"
    filepath = pathlib.Path(output_folder, filename)
    with filepath.open("w") as fh:
        fh.write(response)


# TODO test.
def save_genbank_file(seqrecord, accession, name, output_folder):
    """Save retrieved record to file."""
    filename = f"{name}__{accession}.txt"
    filepath = pathlib.Path(output_folder, filename)
    SeqIO.write(seqrecord, str(filepath), "genbank")


# TODO unittest.
def set_phagesdb_gnm_date(gnm):
    """Set the date of a PhagesDB genome object."""

    # Since there may be multiple 'dates' associated with a genome in PhagesDB,
    # this date attribute can't be directly set when parsing the data.
    # Some phages may have a file but no associated
    # date tagged with that file (since date tagging has not
    # always been implemented). If there is a date, it is
    # formatted as: '2017-02-15T10:37:21Z'. If there is no date,
    # it is Null, but change this to a standarized 'empty' value of 1/1/0001.
    date = gnm.misc["qced_genbank_file_date"]
    if date is None:
        # The date used to be set to default (1/1/0001), but the force flag
        # now relies on this early date, so here set to a date
        # more recent than default.
        # gnm.date = constants.EMPTY_DATE
        gnm.date = datetime.strptime("1/1/1000 00:00:00", "%m/%d/%Y %H:%M:%S")
    else:
        #e.g. 2018-09-01T22:19:33-04:00 (string)
        date = date.split("T")[0]
        #e.g. 2018-09-01 (string)
        gnm.date = datetime.strptime(date, "%Y-%m-%d")
        #e.g. 2018-09-01 00:00:00 (datetime.datetime)

# TODO unittest.
def set_phagesdb_gnm_file(gnm):
    """Set the filename of a PhagesDB genome object."""
    # Since there may be multiple 'files'' associated with a genome in PhagesDB,
    # the set filename is not necessarily for the flat file.
    file_url = gnm.misc["qced_genbank_file"]
    if file_url is None:
        gnm.filename = ""
    else:
        gnm.filename = file_url


# TODO unittest.
def get_genbank_data(output_folder, genome_dict, ncbi_cred_dict={},
                     genbank_results=False, force=False):
    """Run sub-pipeline to retrieve genomes from GenBank."""
    # Flow of the NCBI record retrieval process:
    # 1 Create list of phages to check for updates at NCBI (completed above)
    # 2 Using esearch, verify which accessions are valid
    # 3 Using esummary, get update date for each valid accession
    # 4 Using efetch, retrieve flat files for NCBI records newer than
    # the MySQL database date
    # 5 Save new records in a folder and create an import table for them

    print(f"\n\nDownloading genome(s) from GenBank.")
    # Create output folder
    ncbi_folder = pathlib.Path(output_folder, GENBANK_FOLDER)
    ncbi_folder.mkdir()
    ncbi_results_list = []

    # Iterate through each phage in the MySQL database
    tup1 = sort_by_accession(genome_dict, force=force)
    ncbi_results_list.extend(tup1[0])
    accession_dict = tup1[1]

    # More setup variables if NCBI updates are desired.  NCBI Bookshelf resource
    # "The E-utilities In-Depth: Parameters, Syntax and More", by Dr. Eric
    # Sayers, recommends that a single request not contain more than about 200
    # UIDS so we will use that as our batch size, and all Entrez requests must
    # include the user's email address and tool name.
    ncbi.set_entrez_credentials(tool=ncbi_cred_dict["tool"],
                                email=ncbi_cred_dict["email"],
                                api_key=ncbi_cred_dict["api_key"])

    results = retrieve_records(accession_dict, ncbi_folder, batch_size=200)
    ncbi_results_list.extend(results)

    # Record retrieval results for all phages.
    if genbank_results == True:
        output_genbank_summary(ncbi_folder, ncbi_results_list)

    # Print summary of script
    tallies = compute_genbank_tallies(ncbi_results_list)
    print_genbank_tallies(tallies)

    # Now remove empty folders.
    if len(basic.identify_contents(ncbi_folder, kind=None)) == 0:
        ncbi_folder.rmdir()


# TODO unittest.
def output_genbank_summary(output_folder, results):
    """Save summary of GenBank retrieval results to file."""
    filepath = pathlib.Path(output_folder, GENBANK_RESULTS_TABLE)
    fileio.export_data_dict(results, filepath, NCBI_RESULTS_COLUMNS,
                           include_headers=True)


# TODO unittest.
def compute_genbank_tallies(results):
    """Tally results from GenBank retrieval."""
    dict = {}
    for s in GENBANK_STATUS:
        dict[s] = 0
    for d in results:
        r = d["result"]
        dict[r] += 1
    dict["total"] = len(results)
    dict["auto_updated"] = dict["total"] - dict["not_auto_updated"]
    dict["accession"] = (dict["auto_updated"] -
                         dict["no_accession"] -
                         dict["duplicate_accession"])
    return dict


# TODO unittest.
def print_genbank_tallies(tallies):
    """Print results of GenBank retrieval."""

    print("\n\n\nSummary of GenBank data retrieval:")
    print("Of the genomes in the MySQL database:")
    print(f"{tallies['total']:>6}: total")
    print(f"{tallies['not_auto_updated']:>6}: not auto-updated")
    print(f"{tallies['auto_updated']:>6}: auto-updated")

    print("\nOf the auto-updated genomes:")
    print(f"{tallies['no_accession']:>6}: no accession")
    print(f"{tallies['duplicate_accession']:>6}: duplicated accession")
    print(f"{tallies['accession']:>6}: unique accession")

    print("\nOf the auto-updated genomes with unique accessions:")
    print(f"{tallies['retrieval_failure']:>6}: could not be retrieved")
    print(f"{tallies['docsum_not_new']:>6}: retrieved but docsum not new")
    print(f"{tallies['record_not_new']:>6}: retrieved but record not new")
    print(f"{tallies['retrieved_for_import']:>6}: retrieved for import")


# TODO unittest.
def sort_by_accession(genome_dict, force=False):
    """Sort genome objects based on their accession status.

    Only retain data if genome is set to be automatically updated,
    there is a valid accession, and the accession is unique.
    """
    results = []
    accession_dict = {}
    check_set = {None, "", "none"}
    # Get a list of duplicated accessions.
    unique, duplicate = create_accession_sets(genome_dict)

    # Iterate through each phage in the MySQL database
    for key in genome_dict.keys():
        status = None
        gnm = genome_dict[key]
        if gnm.retrieve_record != 1 and force == False:
            status = NOT_AUTO
        elif gnm.accession in check_set:
            status = NO_ACC
        elif gnm.accession in duplicate:
            status = DUPE_ACC
        else:
            pass

        if status is None:
            # Dictionary of phage data based on unique accessions
            # Key = accession
            # Value = genome object
            accession_dict[gnm.accession] = gnm
        else:
            result = create_results_dict(gnm, "NA", status)
            results.append(result)

    return (results, accession_dict)


# TODO unittest.
def process_failed_retrieval(accession_list, accession_dict):
    """Create list of dictionaries for records that could not be retrieved."""
    l = []
    for acc in accession_list:
        gnm = accession_dict[acc]
        result = create_results_dict(gnm, "NA", RET_FAIL)
        l.append(result)
    return l


# TODO unittest.
def retrieve_records(accession_dict, ncbi_folder, batch_size=200):
    """Retrieve GenBank records."""
    print("\n\nRetrieving records from NCBI")
    genome_folder = pathlib.Path(ncbi_folder, GENOME_FOLDER)
    genome_folder.mkdir()
    retrieval_errors = []
    results = []
    tickets_list = []
    accessions = list(accession_dict.keys())
    mod_accessions = [accession + "[ACCN]" for accession in accessions]

    # When retrieving in batch sizes, first create the list of values
    # indicating which indices of the accessions should be used
    # to create each batch.
    # For instace, if there are five accessions, batch size of two produces
    # indices = 0,2,4

    batch_indices = basic.create_indices(mod_accessions, batch_size)
    print(f"There are {len(mod_accessions)} GenBank accession(s) to check.")
    for indices in batch_indices:
        start = indices[0]
        stop = indices[1]
        print(f"Checking accessions {start + 1} to {stop}...")
        esearch_term = " | ".join(mod_accessions[start:stop])

        # Use esearch for each accession
        # First use esearch to verify the accessions are valid.
        search_record = ncbi.run_esearch(db="nucleotide", term=esearch_term,
                                         usehistory="y")
        search_count = int(search_record["Count"])
        search_webenv = search_record["WebEnv"]
        search_query_key = search_record["QueryKey"]

        # Keep track of the accessions that failed to be located in NCBI
        # Each accession in the error list is formatted "accession[ACCN]"
        current_batch_size = stop - start
        if search_count < current_batch_size:
            search_failure = search_record["ErrorList"]["PhraseNotFound"]
            for accession in search_failure:
                retrieval_errors.append(accession[:-6])

        # Now get summaries for these records using esummary
        summary_records = ncbi.get_summaries(db="nucleotide",
                                             query_key=search_query_key,
                                             webenv=search_webenv)
        results_tuple = get_accessions_to_retrieve(summary_records,
                                                   accession_dict)
        accessions_to_retrieve = results_tuple[0]
        results.extend(results_tuple[1])

        if len(accessions_to_retrieve) > 0:
            # Use efetch to retrieve the record.
            output_list = ncbi.get_records(accessions_to_retrieve,
                                           db="nucleotide",
                                           rettype="gb",
                                           retmode="text")

            # TODO check_record_date may be redundant. It checks date within the
            # record. Earlier in the pipeline, the docsum date has already been
            # checked though. So if docsum date is identical to date in the
            # record, this is redundant.
            tup = check_record_date(output_list, accession_dict)
            new_record_list = tup[0]
            # list of results dictionaries
            results.extend(tup[1])

            if len(new_record_list) > 0:
                tickets = save_and_tickets(new_record_list, accession_dict,
                                           genome_folder)
                tickets_list.extend(tickets)

    if len(tickets_list) > 0:
        create_ticket_table(tickets_list, ncbi_folder)

    # Remove genome folder if empty.
    if len(basic.identify_contents(genome_folder, kind=None)) == 0:
        genome_folder.rmdir()

    # Report the genomes that could not be retrieved.
    failed = process_failed_retrieval(retrieval_errors, accession_dict)
    results.extend(failed)

    return results


# TODO unittest.
def get_accessions_to_retrieve(summary_records, accession_dict):
    """Review GenBank summary to determine which records are new."""
    accessions = []
    results = []
    for doc_sum in summary_records:
        doc_sum_accession = doc_sum["Caption"]
        doc_sum_date = datetime.strptime(doc_sum["UpdateDate"], "%Y/%m/%d")
        gnm = accession_dict[doc_sum_accession]

        # If Document Summary date is newer than the MySQL database date,
        # or if the genome being evaluated is currently draft status but
        # we were able to retrieve a Genbank record (i.e. it's not a
        # draft anymore), append the accession to the list of accessions
        # from this batch to retrieve.
        # Otherwise, record that the GenBank record isn't new.
        if doc_sum_date > gnm.date or gnm.annotation_status == "draft":
            accessions.append(doc_sum_accession)
        else:
            result = create_results_dict(gnm, doc_sum_date, OLD_DOCSUM)
            results.append(result)
    return (accessions, results)


# TODO unittest.
def check_record_date(record_list, accession_dict):
    """Check whether the GenBank record is new."""
    results = []
    new_record_list = []
    # For any records that were retrieved, mark their data in the NCBI
    # results file, and save their records as Genbank files, and create
    # import table entries for them.
    for index in range(len(record_list)):
        record = record_list[index]
        date = record.annotations["date"]
        date = datetime.strptime(date, '%d-%b-%Y')
        accession = record.annotations["accessions"][0]
        # accession = record.name
        # accession = accession.split('.')[0]
        gnm = accession_dict[accession]

        # Save new records in a folder and create an import table row
        # for them. If the genome is currently a draft annotation, create
        # an import ticket for replacement regardless of the date in
        # the Genbank record. This ensures that if a user fails to
        # upload a manual annotation to PhagesDB, once the Genbank
        # accession becomes active MySQL database will get the new version.
        # This should always happen, since we've only retrieved new records.
        if (date > gnm.date or gnm.annotation_status == "draft"):
            data_dict = create_results_dict(gnm, date, RETRIEVED)
            results.append(data_dict)
            new_record_list.append(record)
            # Determine what the status of the genome will be.
            # If the genome in the MySQL database was already 'final' or 'unknown',
            # then keep the status unchanged. If the genome in the MySQL database
            # was 'draft', the status should be changed to 'final'.
            if gnm.annotation_status == "draft":
                gnm.annotation_status = "final"
        else:
            data_dict = create_results_dict(gnm, date, OLD_RECORD)
            results.append(data_dict)

    return (new_record_list, results)


# TODO unittest.
def save_and_tickets(record_list, accession_dict, output_folder):
    """Save flat files retrieved from GenBank and create import tickets."""
    # Both tasks accomplished at the same time to
    # iterate through record list once.
    l = []
    for record in record_list:
        # accession = seqrecord.name
        # accession = accession.split('.')[0]
        accession = record.annotations["accessions"][0]
        gnm = accession_dict[accession]
        save_genbank_file(record, accession, gnm.name, output_folder)
        tkt = create_genbank_ticket(gnm)
        l.append(tkt)
    return l

# TODO test.
def create_ticket_table(tickets, output_folder):
    """Save tickets associated with retrieved from GenBank files."""
    filepath = pathlib.Path(output_folder, IMPORT_TABLE)
    tickets = convert_tickets_to_dict(tickets)
    fileio.export_data_dict(tickets, filepath, IMPORT_COLUMNS,
                           include_headers=True)


# TODO unittest.
def create_accession_sets(genome_dict):
    """Generate set of unique and non-unique accessions.

    Input is a dictionary of pdm_utils genome objects."""
    l = []
    for id in genome_dict.keys():
        gnm = genome_dict[id]
        # In the MySQL databse, empty accession data is stored as ''.
        # There are no NULL accessions.
        if gnm.accession != "":
            l.append(gnm.accession)
    unique, duplicated = basic.identify_unique_items(l)

    if len(duplicated) > 0:
        print("There are duplicated accessions. Some data will not be "
             "retrieved from GenBank:")
        for duplicate in duplicated:
            print(duplicate)
        input("\n\nPress ENTER to continue.")

    return unique, duplicated


# TODO unittest.
def create_results_dict(gnm, genbank_date, result):
    """Create a dictionary of data summarizing NCBI retrieval status."""
    d = {
        "phage_id": gnm.id,
        "phage_name": gnm.name,
        "accession": gnm.accession,
        "annotation_status": gnm.annotation_status,
        "mysql_date": gnm.date,
        "genbank_date": genbank_date,
        "result": result
        }
    return d


# TODO unittest.
def get_draft_data(output_path, phage_id_set):
    """Run sub-pipeline to retrieve auto-annotated 'draft' genomes."""

    if len(phage_id_set) > 0:
        print(f"\n\nRetrieving {len(phage_id_set)} genome(s) from PECAAN")
        phage_id_list = list(phage_id_set)
        pecaan_folder = pathlib.Path(output_path, PECAAN_FOLDER)
        pecaan_folder.mkdir()
        retrieve_drafts(pecaan_folder, phage_id_list)
    else:
        print("No new genomes to retrieve from PECAAN.")


# TODO unittest.
def retrieve_drafts(output_folder, phage_list):
    """Retrieve auto-annotated 'draft' genomes from PECAAN."""

    genome_folder = pathlib.Path(output_folder, GENOME_FOLDER)
    genome_folder.mkdir()
    failed = []
    tickets = []

    # Iterate through each row in the file
    for new_phage in phage_list:
        pecaan_link = constants.PECAAN_PREFIX + new_phage
        response = phagesdb.retrieve_url_data(pecaan_link)
        if response == "":
            print(f"Error: unable to retrieve {new_phage} draft genome.")
            print(pecaan_link)
            failed.append(new_phage)
        else:
            save_pecaan_file(response, new_phage, genome_folder)
            tkt = create_draft_ticket(new_phage)
            tickets.append(tkt)
            print(f"{new_phage} retrieved from PECAAN.")

    if len(tickets) > 0:
        create_ticket_table(tickets, output_folder)
        print(f"{len(tickets)} phage(s) were successfully retrieved")

    if len(failed) > 0:
        print(f"{len(failed)} phage(s) failed to be retrieved:")
        for item in failed:
            print(item)
        input("\n\nPress ENTER to continue.")


# TODO test.
def create_draft_ticket(name):
    """Create ImportTicket for draft genome."""
    tkt = ticket.ImportTicket()
    tkt.type = "add"
    tkt.phage_id = name
    tkt.description_field = "product"
    tkt.eval_mode = "draft"
    tkt.data_dict = {
        "host_genus": "retrieve",
        "cluster": "retrieve",
        "subcluster": "retrieve",
        "annotation_status": "draft",
        "annotation_author": 1,
        "accession": "none",
        "retrieve_record": 1
        }
    return tkt


# TODO test.
def create_phagesdb_ticket(phage_id):
    """Create ImportTicket for PhagesDB genome."""

    # Since the PhagesDB phage has been matched to
    # the MySQL database phage, the AnnotationAuthor field
    # could be assigned from the current mysqldb author
    # variable. However, since this genbank-formatted
    # file is acquired through PhagesDB, both the
    # Annotation status is expected to be 'final' and
    # the Annotation author is expected to be 'hatfull'.
    tkt = ticket.ImportTicket()
    tkt.type = "replace"
    tkt.phage_id = phage_id
    tkt.description_field = "product"
    tkt.eval_mode = "final"
    tkt.data_dict = {
        "host_genus": "retain", # formerly "retrieve",
        "cluster": "retain", # formerly "retrieve",
        "subcluster": "retain", # formerly "retrieve",
        "annotation_status": "final",
        "annotation_author": 1,
        "accession": "retain", # formerly "retrieve",
        "retrieve_record": 1
        }
    return tkt


# TODO test.
def create_genbank_ticket(gnm):
    """Create ImportTicket for GenBank record."""
    # Accession is set to 'parse' to ensure that during import,
    # the file's accession is directly compared to the database
    # record's accession.
    tkt = ticket.ImportTicket()
    tkt.type = "replace"
    tkt.phage_id = gnm.id
    tkt.description_field = "product"
    tkt.eval_mode = "auto"
    tkt.data_dict = {
        "host_genus": "retain", # formerly gnm.host_genus,
        "cluster": "retain", # formerly gnm.cluster,
        "subcluster": "retain", # formerly gnm.subcluster,
        "annotation_status": "retain", # formerly gnm.annotation_status,
        "annotation_author": "retain", # formerly gnm.annotation_author,
        "accession": "parse",
        "retrieve_record": "retain", # formerly 1
        }
    return tkt



"""
Functions to multithread process a list of inputs.
"""

from queue import Queue
import time
import threading

from pdm_utils.classes.progress import ProgressBar, show_progress


class MixedThread(threading.Thread):
    def __init__(self, thread_id, work_queue, work_lock,
                 result_queue, result_lock, lock_timeout=-1,
                 kill_sig=None):
        threading.Thread.__init__(self)

        self.name = thread_id

        # Queue filled with tuples containing target function args
        self.work_queue = work_queue
        self.result_queue = result_queue
        # Lock controlling access to queue accross multiple threads
        self.work_lock = work_lock
        self.result_lock = result_lock

        self.lock_timeout = lock_timeout
        self.kill_sig = kill_sig

    def run(self):
        while not self.work_queue.empty():
            # Tell other threads "I'm accessing the queue right now"
            self.work_lock.acquire(timeout=self.lock_timeout)

            job = self.work_queue.get()
            # Let other threads have a turn
            self.work_lock.release()

            target = job[0]
            work_item = job[1]

            # Run target function with args retrieved from queue
            result = target(*work_item)

            # Repeat with the result queue/lock
            self.result_lock.acquire(timeout=self.lock_timeout)
            self.result_queue.put(result)
            self.result_lock.release()

            if self.kill_sig is not None:
                if self.kill_sig():
                    break


def create_threads(work_queue, work_lock, result_queue, result_lock,
                   num_threads, lock_timeout=-1, kill_sig=None):
    """Function to create a pseudo-MixIn Thread class for multithreading.

    :param target: Function used by the thread to process work items.
    :type target: Function
    :param work_queue: Queue containing tuples of target function args.
    :type work_queue: Queue
    :param work_lock: Lock for controlling access of work queue accross threads
    :type work_lock: Lock
    :param result_queue: Queue to place results into
    :type result_queue: Queue
    :param result_lock: Lock for controlling result queue accross threads.
    :type result_lock: Lock
    :param num_threads: Number of threads to be created to process work stack.
    :type num_threads: int
    :param lock_timeout: Amount of time thread should block during lock.acquire
    :type lock_timeout: int
    :param kill_sig: Function that returns a boolean that controls lock killing
    :type kill_sig: Callable
    :returns: Returns a list of MixIn Thread objects with target function
    :rtype: list[Thread]
    """
    threads = []
    for x in range(num_threads):
        threads.append(MixedThread(x+1, work_queue, work_lock,
                                   result_queue, result_lock,
                                   lock_timeout=lock_timeout, kill_sig=None))

    return threads


def multithread(work_items, threads, target, verbose=False,
                lock_timeout=-1, join_timeout=None):
    """Runs list of work items with specified number of threads.

    :param target: Function used by the thread to process work items.
    :type target: Function
    :type work_items: List containing tuples of target function args.
    :type work_stack: list
    :param threads: Number of threads to be created to process work_stack.
    :type threads: int
    """
    # Create lock which prevents work queue access deadlocks
    work_lock = threading.Lock()
    work_queue = Queue()

    # Create lock which prevents result queue access deadlocks
    result_lock = threading.Lock()
    result_queue = Queue()

    tasks = 0
    if verbose:
        interval = max([1, len(work_items) // 100])

        for i in range(len(work_items)):
            if i % interval == 0:
                tasks += 1
                work_queue.put((show_progress, (i, len(work_items))))

            tasks += 1
            work_queue.put((target, work_items[i]))

        tasks += 1
        work_queue.put((show_progress, (len(work_items), len(work_items))))
    else:
        for item in work_items:
            tasks += 1
            work_queue.put((target, item))

    kill_threads = False
    kill_sig = (lambda: kill_threads)
    threads = create_threads(work_queue, work_lock, result_queue, result_lock,
                             threads, lock_timeout=lock_timeout,
                             kill_sig=kill_sig)

    for thread in threads:
        # Calls thread run()
        thread.start()

    while not work_queue.empty():
        time.sleep(0.25)

    for thread in threads:
        thread.join(timeout=join_timeout)

    if join_timeout is not None:
        for _ in range(join_timeout):
            threads_alive = False
            for thread in threads:
                if not threads_alive:
                    threads_alive = thread.is_alive()

            if not threads_alive:
                break

            time.sleep(1)

    kill_threads = True

    # Remove non-Progress results
    results = []
    for i in range(tasks):
        result = result_queue.get()
        if result is not None and not isinstance(result, ProgressBar):
            results.append(result)

    return results



"""Pipeline to review discrepant or outdated cds product annotations. """

import argparse
import sys
import textwrap
import time
from pathlib import Path

from pdm_utils.functions import annotation
from pdm_utils.functions import basic
from pdm_utils.functions import configfile
from pdm_utils.functions import fileio
from pdm_utils.functions import mysqldb_basic
from pdm_utils.functions import pipelines_basic
from pdm_utils.functions import querying

# -----------------------------------------------------------------------------
# GLOBAL VARIABLES

DEFAULT_FOLDER_NAME = f"{time.strftime('%Y%m%d')}_pham_review"

REVIEW_HEADER = ["Pham", "Final Call", "#Members", "Clusters", "#Functions",
                 "Functional Calls"]
GR_HEADER = ["Gene",
             "Phage",
             "Gene#",
             "Cluster",
             "Subcluster",
             "Functional Call",
             "Translation"]

REVIEW_DATA_COLUMNS = ["gene.GeneID", "phage.Cluster"]
GR_DATA_COLUMNS = ["phage.PhageID", "gene.Name", "phage.Cluster",
                   "phage.Subcluster", "gene.Notes", "gene.Translation"]

BASE_CONDITIONALS = ("phage.Status = final AND "
                     "phage.AnnotationAuthor = 1 AND "
                     "phage.RetrieveRecord = 1")

# -----------------------------------------------------------------------------
# MAIN FUNCTIONS


def main(unparsed_args_list):
    """Uses parsed args to run the entirety of the pham_review pipeline.

    :param unparsed_args_list: Input a list of command line args.
    :type unparsed_args_list: list[str]
    """
    args = parse_pham_review(unparsed_args_list)

    config = configfile.build_complete_config(args.config_file)

    alchemist = pipelines_basic.build_alchemist(args.database, config=config)

    values = pipelines_basic.parse_value_input(args.input)

    if not args.all_reports:
        gr_reports = args.gene_reports
        s_report = args.summary_report
        psr_reports = args.pham_summary_reports
    else:
        gr_reports = True
        s_report = True
        psr_reports = True

    execute_pham_review(
                   alchemist,
                   folder_path=args.folder_path, folder_name=args.folder_name,
                   no_review=args.no_review, values=values, force=args.force,
                   filters=args.filters, groups=args.groups, sort=args.sort,
                   s_report=s_report, gr_reports=gr_reports,
                   production=args.production, psr_reports=psr_reports,
                   verbose=args.verbose)


def parse_pham_review(unparsed_args_list):
    """Parses review arguments and stores them with an argparse object.

    :param unparsed_args_list: Input a list of command line args.
    :type unparsed_args_list: list[str]
    :returns: ArgParse module parsed args.
    """
    DATABASE_HELP = """
        Name of the MySQL database to export from.
        """

    FORCE_HELP = """
        Review option that aggresively creates and overwrites directories.
        """
    CONFIG_FILE_HELP = """
        Review option that enables use of a config file for sourcing
        credentials
            Follow selection argument with the path to the config file
            specifying MySQL and NCBI credentials.
        """
    VERBOSE_HELP = """
        Review option that enables progress print statements.
        """
    FOLDER_PATH_HELP = """
        Review option to change the path
        of the directory where the exported files are stored.
            Follow selection argument with the path to the
            desired directory.
        """
    FOLDER_NAME_HELP = """
        Review option to change the name
        of the directory where the exported files are stored.
            Follow selection argument with the desired name.
        """

    ALL_REPORTS_HELP = """
        Review option to toggle export of all report options.
        """
    GENE_REPORTS_HELP = """
        Review option to toggle export of supplemental information about
        the genes in the phams selected.
        """
    SUMMARY_REPORT_HELP = """
        Review option to toggle export of supplemental information about
        the profile of the phages selected.
        """
    PHAM_SUMMARY_REPORT_HELP = """
        Review option to toggle export of supplemental information about
        the profile of a pham selected for review.
        """
    PRODUCTION_HELP = """
        Review option to toggle additional filters to support production-level
        review.
        """

    REVIEW_HELP = """
        Review option to toggle review of phams.  If enabled,
        phams inputted or auto-generated will not be reviewed
        for inconsistencies.
        """
    IMPORT_FILE_HELP = """
        Selection input option that imports values from a csv file.
            Follow selection argument with path to the
            csv file containing the names of each genome in the first column.
        """
    IMPORT_NAMES_HELP = """
        Selection input option that imports values from cmd line input.
            Follow selection argument with space separated
            names of genomes in the database.
        """

    WHERE_HELP = """
        Data filtering option that filters data by the inputted expressions.
            Follow selection argument with formatted filter expression:
                {Table}.{Column}={Value}
        """
    GROUP_BY_HELP = """
        Data selection option that groups data by the inputted columns.
            Follow selection argument with formatted column expressions:
                {Table}.{Column}={Value}
        """
    ORDER_BY_HELP = """
        Data selection option that sorts data by the inputted columns.
            Follow selection argument with formatted column expressions:
                {Table}.{Column}={Value}
        """

    parser = argparse.ArgumentParser()

    parser.add_argument("database", type=str,  help=DATABASE_HELP)

    parser.add_argument("-f", "--force", action="store_true",
                        help=FORCE_HELP)
    parser.add_argument("-c", "--config_file",
                        type=pipelines_basic.convert_file_path,
                        help=CONFIG_FILE_HELP)
    parser.add_argument("-m", "--folder_name",
                        type=str,  help=FOLDER_NAME_HELP)
    parser.add_argument("-o", "--folder_path", type=Path,
                        help=FOLDER_PATH_HELP)
    parser.add_argument("-v", "--verbose", action="store_true",
                        help=VERBOSE_HELP)

    parser.add_argument("-a", "--all_reports", action="store_true",
                        help=ALL_REPORTS_HELP)
    parser.add_argument("-gr", "--gene_reports", action="store_true",
                        help=GENE_REPORTS_HELP)
    parser.add_argument("-sr", "--summary_report", action="store_true",
                        help=SUMMARY_REPORT_HELP)
    parser.add_argument("-psr", "--pham_summary_reports", action="store_true",
                        help=PHAM_SUMMARY_REPORT_HELP)
    parser.add_argument("-p", "--production", action="store_true",
                        help=PRODUCTION_HELP)

    parser.add_argument("-nr", "--no_review", action="store_true",
                        help=REVIEW_HELP)
    parser.add_argument("-if", "--import_files", dest="input",
                        type=pipelines_basic.convert_file_path,
                        help=IMPORT_FILE_HELP)
    parser.add_argument("-in", "--import_names", nargs="*", dest="input",
                        help=IMPORT_NAMES_HELP)

    parser.add_argument("-w", "--where", nargs="?", dest="filters",
                        help=WHERE_HELP)
    parser.add_argument("-g", "--group_by", nargs="*", dest="groups",
                        help=GROUP_BY_HELP)
    parser.add_argument("-s", "--order_by", nargs="*", dest="sort",
                        help=ORDER_BY_HELP)

    default_folder_name = DEFAULT_FOLDER_NAME

    parser.set_defaults(folder_name=default_folder_name,
                        folder_path=None,
                        input=[], filters="", groups=[], sort=[],
                        config_file=None,
                        no_review=False, gene_report=False,
                        summary_report=False, verbose=False)

    parsed_args = parser.parse_args(unparsed_args_list[2:])
    return parsed_args


def execute_pham_review(
                   alchemist, folder_path=None,
                   folder_name=DEFAULT_FOLDER_NAME, no_review=False, values=[],
                   filters="", groups=[], sort=[], s_report=False,
                   gr_reports=False, psr_reports=False, production=False,
                   verbose=False, force=False):
    """Executes the entirety of the pham review pipeline.

    :param alchemist: A connected and fully built AlchemyHandler object.
    :type alchemist: AlchemyHandler
    :param folder_path: Path to a valid dir for new dir creation.
    :type folder_path: Path
    :param folder_name: A name for the export folder.
    :type folder_name: str
    :param csv_title: Title for an appended csv file prefix.
    :type csv_title: str
    :param review: A boolean to toggle filtering of phams by pham discrepancies
    :type review: bool
    :param values: List of values to filter database results.
    :type values: list[str]
    :param force: A boolean to toggle aggresive building of directories.
    :type force: bool
    :param filters: A list of lists with filter values, grouped by ORs.
    :type filters: list[list[str]]
    :param groups: A list of supported MySQL column names to group by.
    :type groups: list[str]
    :param sort: A list of supported MySQL column names to sort by.
    :param gr_reports: A boolean to toggle export of additional pham info
    :type gr_reports: bool
    :param production: Toggles additional filters for production-level review
    :type production: bool
    :param verbose: A boolean value to toggle progress print statements.
    :type verbose: bool
    """
    db_filter = pipelines_basic.build_filter(alchemist, "gene.PhamID", filters,
                                                        values=values,
                                                        verbose=verbose)
    if production:
        db_filter.add(BASE_CONDITIONALS)
        db_filter.update()
    else:
        conditionals = db_filter.build_where_clauses()
        db_filter.values = db_filter.build_values(where=conditionals)

    if not db_filter.values:
        print("Current settings produced no database hits.")
        sys.exit(1)
    else:
        if verbose:
            print(f"Identified {db_filter.hits()} phams to review...")

    if not no_review:
        review_phams(db_filter, verbose=verbose)

    if sort:
        db_filter.sort(sort)

    if verbose:
        print("Creating export folder...")
    export_path = pipelines_basic.create_working_path(folder_path, folder_name,
                                                      force=force)

    conditionals_map = pipelines_basic.build_groups_map(db_filter, export_path,
                                                        groups=groups,
                                                        verbose=verbose,
                                                        force=force)

    if verbose:
        print("Prepared query and path structure, beginning review export...")
    original_phams = db_filter.values
    gr_data_cache = {}
    psr_data_cache = {}
    for mapped_path in conditionals_map.keys():
        conditionals = conditionals_map[mapped_path]
        db_filter.values = original_phams
        db_filter.values = db_filter.build_values(where=conditionals)

        pipelines_basic.create_working_dir(mapped_path, force=force)

        review_data = get_review_data(alchemist, db_filter, verbose=verbose)
        write_report(review_data, mapped_path, REVIEW_HEADER,
                     csv_name="FunctionReport", verbose=verbose)

        if s_report:
            summary_data = get_summary_data(alchemist, db_filter)
            write_summary_report(alchemist, summary_data, mapped_path,
                                 verbose=verbose)

        if gr_reports or psr_reports:
            execute_pham_report_export(alchemist, db_filter, mapped_path,
                                       gr_reports=gr_reports,
                                       psr_reports=psr_reports,
                                       gr_data_cache=gr_data_cache,
                                       psr_data_cache=psr_data_cache,
                                       verbose=verbose)


def execute_pham_report_export(alchemist, db_filter, export_path,
                               gr_reports=False, gr_data_cache={},
                               psr_reports=False, psr_data_cache={},
                               verbose=False):
    """Executes export of gene data for a reviewed pham.

    :param alchemist: A connected and fully built AlchemyHandler object.
    :type alchemist: AlchemyHandler
    :param export_path: Path to a valid dir for new file creation.
    :type export_path: Path
    :param gr_data_cache: Total data extracted for gene reports.
    :type gr_data_cache: dict
    :param verbose: A boolean value to toggle progress print statements.
    :type verbose: bool
    """
    phams = db_filter.values

    pham_report_path = export_path.joinpath("PhamReports")
    pham_report_path.mkdir()

    for pham in phams:
        pham_path = pham_report_path.joinpath(str(pham))
        pham_path.mkdir()

        db_filter.values = [pham]

        if gr_reports:
            gr_data = gr_data_cache.get(pham)
            if gr_data is None:
                gr_data = get_gr_data(alchemist, db_filter, verbose=verbose)
                gr_data_cache[pham] = gr_data

            write_report(gr_data, pham_path, GR_HEADER,
                         csv_name=f"{pham}_GeneReport",
                         verbose=verbose)
        if psr_reports:
            psr_data = psr_data_cache.get(pham)
            if psr_data is None:
                psr_data = get_psr_data(alchemist, db_filter, verbose=verbose)
                psr_data_cache[pham] = psr_data

            write_pham_summary_report(psr_data, pham_path, verbose=verbose)


def review_phams(db_filter, verbose=False):
    """Finds and stores phams with discrepant function calls in a Filter.

    """
    notes = db_filter.get_column("gene.Notes")

    if verbose:
        print("Reviewing phams...")

    reviewed_phams = []
    for index in range(len(db_filter.values)):
        pham = db_filter.values[index]
        if verbose:
            print(f"...Analyzing Pham {pham}...")

        query = querying.build_count(db_filter.graph, notes.distinct(),
                                     where=(db_filter.key == pham))
        func_count = mysqldb_basic.scalar(db_filter.engine, query)

        if func_count <= 1:
            continue

        if verbose:
            print(f"......Detected discrepencies in Pham {pham}")
        reviewed_phams.append(pham)

    if verbose:
        print(f"Detected {len(reviewed_phams)} disrepent phams...")

    db_filter.values = reviewed_phams


def write_report(data, export_path, header, csv_name="Report", verbose=False):
    """Outputs a csv file
    """
    if not export_path.is_dir():
        print("Passed in path is not a directory.")
        sys.exit(1)

    file_path = export_path.joinpath(f"{csv_name}.csv")
    if verbose:
        print(f"Writing {file_path.name} in {export_path.name}...")

    fileio.export_data_dict(data, file_path, header, include_headers=True)


def write_summary_report(alchemist, summary_data, export_path, verbose=False):
    if verbose:
        print(f"Writing SummaryReport.txt in {export_path.name}")

    s_path = export_path.joinpath("SummaryReport.txt")
    s_file = open(s_path, "w")

    version_data = summary_data["version_data"]
    recurring_phages = summary_data["recurring_phages"]
    recent_phages = summary_data["recent_phages"]

    s_file.write(f"Phams reviewed on: {time.strftime('%d-%m-%Y')}\n")
    s_file.write(f"Database reviewed: {alchemist.database}\n")
    s_file.write(f"Schema version: {version_data['SchemaVersion']} "
                 f"Database version: {version_data['Version']}\n\n")
    s_file.write("Phams reviewed using the following base conditionals:\n")
    s_file.write(f"    {BASE_CONDITIONALS}\n")

    s_file.write("\n\n")
    s_file.write(f"Most occuring phages: {', '.join(recurring_phages)}\n")
    s_file.write(f"Phages recently submitted: {', '.join(recent_phages)}\n")
    s_file.close()


def write_pham_summary_report(psr_data, export_path, verbose=False):
    if verbose:
        print(f"Writing SummaryReport.txt in {export_path.name}")

    s_path = export_path.joinpath("SummaryReport.txt")
    s_file = open(s_path, "w")

    s_file.write(f"Pham: {export_path.name}\n\n")

    s_file.write("Annotations of genes to the left:\n")
    s_file.write(f"{psr_data['left_annotations']}\n")

    s_file.write("\n")
    s_file.write("Annotations of genes to the right:\n")
    s_file.write(f"{psr_data['right_annotations']}\n")

    s_file.write("\n\n")
    s_file.write("Conserved Database Domains in pham:\n\n")
    for domain in psr_data["cdd_domains"]:
        s_file.write(f"{domain}\n\n")

    s_file.close()
# -----------------------------------------------------------------------------
# REVIEW-SPECIFIC HELPER FUNCTIONS


# TODO Documentation
def get_review_data(alchemist, db_filter, verbose=False):
    """
    """
    if verbose:
        print("Retrieving data for phams...")

    review_columns = get_review_data_columns(alchemist)
    row_dicts = db_filter.retrieve(review_columns)

    review_data = []
    for pham in row_dicts.keys():
        if verbose:
            print(f"...Processing data for pham {pham}...")
        row_dict = row_dicts[pham]
        row_dict["Notes"] = annotation.get_count_annotations_in_pham(
                                                            alchemist, pham)

        format_review_data(row_dict, pham)
        review_data.append(row_dict)

    return review_data


def get_summary_data(alchemist, db_filter, verbose=False):
    phams = db_filter.values

    phages_data = db_filter.retrieve("phage.PhageID", filter=True)

    db_filter.values = phams
    db_filter.transpose("phage.PhageID", set_values=True)
    db_filter.sort("phage.DateLastModified")

    version_data = mysqldb_basic.get_first_row_data(alchemist.engine,
                                                    "version")

    summary_data = {}
    summary_data["recent_phages"] = db_filter.values
    summary_data["recurring_phages"] = phages_data
    summary_data["version_data"] = version_data

    format_summary_data(summary_data)

    db_filter.values = phams
    db_filter.key = "gene.PhamID"
    return summary_data


def get_gr_data(alchemist, db_filter, verbose=False):
    pham = db_filter.values[0]
    if verbose:
        print(f"Retrieving genes in pham {pham}...")
    db_filter.transpose("gene.GeneID", set_values=True)

    pg_columns = get_gr_data_columns(alchemist)
    row_dicts = db_filter.retrieve(pg_columns)

    gr_data = []
    for gene in row_dicts.keys():
        if verbose:
            print(f"...Processing data for gene {gene}...")

        row_dict = row_dicts[gene]

        format_gr_data(row_dict, gene)
        gr_data.append(row_dict)

    db_filter.values = [pham]
    db_filter.key = "gene.PhamID"
    return gr_data


def get_psr_data(alchemist, db_filter, verbose=False):
    pham = db_filter.values[0]
    adjacent_genes = annotation.get_genes_adjacent_to_pham(alchemist, pham)
    psr_data = {}

    psr_data["left_annotations"] = annotation.get_count_annotations_in_genes(
                                                alchemist, adjacent_genes[0])
    psr_data["right_annotations"] = annotation.get_count_annotations_in_genes(
                                                alchemist, adjacent_genes[1])

    db_filter.values = [pham]
    db_filter.transpose("domain.Name", set_values=True)
    cdd_domains = db_filter.retrieve("domain.Description")
    psr_data["cdd_domains"] = cdd_domains

    format_psr_data(psr_data)

    db_filter.values = [pham]
    db_filter.key = "gene.PhamID"
    return psr_data


def format_review_data(row_dict, pham):
    """Function to format function report dictionary keys.

    :param row_dict: Data dictionary for a function report.
    :type row_dict: dict
    :param pham: PhamID to append to the function report data dictionary.
    :type pham: int
    """
    row_dict["Pham"] = pham
    row_dict["Final Call"] = ""
    row_dict["#Members"] = len(row_dict.pop("GeneID"))
    row_dict["#Functions"] = len(row_dict["Notes"])

    row_dict["Clusters"] = ";".join([str(cluster)
                                    for cluster in row_dict.pop("Cluster")])

    count_functional_calls = []
    notes = row_dict.pop("Notes")
    notes = basic.sort_histogram(notes)
    for key in notes.keys():
        if key is None or key == "":
            function = "hypothetical protein"
        else:
            function = key

        count_functional_calls.append(
            "".join([function, f"({str(notes[key])})"]))

    row_dict["Functional Calls"] = ";".join(count_functional_calls)


def format_summary_data(summary_data):
    recent_phages = summary_data["recent_phages"]
    recent_phages.reverse()
    recent_phages = basic.partition_list(recent_phages, 5)[0]
    summary_data["recent_phages"] = recent_phages

    phages_data = summary_data["recurring_phages"]
    phages_histogram = {}
    for pham in phages_data.keys():
        basic.increment_histogram(phages_data[pham]["PhageID"],
                                  phages_histogram)

    recurring_phages = basic.sort_histogram_keys(phages_histogram)
    recurring_phages = basic.partition_list(recurring_phages, 5)[0]
    for i in range(len(recurring_phages)):
        recurring_phages[i] = "".join(
                           [recurring_phages[i],
                            f"({str(phages_histogram[recurring_phages[i]])})"])
    summary_data["recurring_phages"] = recurring_phages


def format_gr_data(row_dict, gene):
    """Function to format gene report dictionary keys.

    :param row_dict: Data dictionary for a gene report.
    :type row_dict: dict
    :param gene: GeneID to append to the gene report data dictionary.
    :type gene: str
    """
    row_dict["Gene"] = gene
    row_dict["Phage"] = row_dict.pop("PhageID")[0]
    row_dict["Gene#"] = row_dict.pop("Name")[0]
    row_dict["Cluster"] = row_dict["Cluster"][0]
    row_dict["Subcluster"] = row_dict["Subcluster"][0]
    row_dict["Translation"] = row_dict["Translation"][0]

    note = row_dict.pop("Notes")[0]
    if note is None or note == "":
        note = "hypothetical protein"

    row_dict["Functional Call"] = note


def format_psr_data(psr_data):
    left_annotations = format_adjacent_annotations(
                                                psr_data["left_annotations"])
    psr_data["left_annotations"] = left_annotations

    right_annotations = format_adjacent_annotations(
                                                psr_data["right_annotations"])
    psr_data["right_annotations"] = right_annotations

    cdd_domains_data = psr_data["cdd_domains"]
    cdd_domains = []
    for domain in cdd_domains_data.keys():
        description = cdd_domains_data[domain]["Description"][0]
        description = textwrap.wrap(description, 72)
        description = "\n".join(description)
        description = textwrap.indent(description, "    ")
        cdd_domain = "".join([domain, "\n", description])
        cdd_domains.append(cdd_domain)
    psr_data["cdd_domains"] = cdd_domains


def format_adjacent_annotations(annotation_histogram):
    annotation_histogram = basic.sort_histogram(annotation_histogram)

    annotation_list = []
    for note in annotation_histogram.keys():
        count = annotation_histogram[note]
        if (note == "") or (note is None):
            note = "hypothetical protein"
        annotation_list.append("".join([note, "(", str(count), ")"]))

    annotations = ", ".join(annotation_list)
    wrapped_annotations = textwrap.wrap(annotations, 72)
    annotations = "\n".join(wrapped_annotations)
    annotations = textwrap.indent(annotations, "    ")
    return annotations


def get_review_data_columns(alchemist):
    """Gets labelled columns for pham function data retrieval.

    :returns: List of labelled columns for function data retrieval.
    :rtype: list[Column]
    """
    review_columns = []

    for column_name in REVIEW_DATA_COLUMNS:
        review_columns.append(querying.get_column(alchemist.metadata,
                                                  column_name))

    return review_columns


def get_gr_data_columns(alchemist):
    """Gets labelled columns for pham gene data retrieval.

    :returns: List of labelled columns for gene data retrieval.
    :rtype: list[Column]
    """
    pg_columns = []

    for column_name in GR_DATA_COLUMNS:
        pg_columns.append(querying.get_column(alchemist.metadata, column_name))

    return pg_columns


if __name__ == "__main__":
    args = sys.argv
    args.insert(0, "")
    main(args)



"""Program to group related gene products into phamilies using MMseqs2
for similarity search and clustering."""

import argparse
from datetime import datetime
import pathlib

from phamerate.__main__ import assemble_phams
from phamerate.SequenceDB import SequenceDB
from phamerate.multiprocess import CPUS

from pdm_utils.classes.alchemyhandler import AlchemyHandler
from pdm_utils.functions.configfile import *
from pdm_utils.functions.phameration import *
from pdm_utils.functions.multiprocess import *

DATE = datetime.now().strftime("%d_%b_%Y")
OUT_DIR = pathlib.Path().cwd().joinpath(f"phamerate__{DATE}")

# Sequence-sequence
SEQ_MIN_SEQ_ID = 35.0
SEQ_COVERAGE = 80.0
SEQ_EVALUE = 0.001
SEQ_SENSITIVITY = 7
SEQ_CLUSTER_MODE = 0
SEQ_CLUSTER_STEPS = 1

HMM_MIN_SEQ_ID = 15.0
HMM_COVERAGE = 70.0
HMM_EVALUE = 0.001
HMM_SENSITIVITY = 7
HMM_CLUSTER_MODE = 0
HMM_CLUSTER_STEPS = 3

EPILOG = """
Steinegger M. and Söding J. MMseqs2 enables sensitive protein
sequence searching for the analysis of massive data sets. Nature
Biotechnology, 2017. doi: 10.1038/nbt.3988"""

DB_SUMMARY = """
Initial database summary:
=============================
 {} total phams
 {} orphams
 {} genes in phams
 {} genes not in phams
 {} total genes
 {} non-redundant genes
=============================
"""


def setup_argparser():
    """Parse command line arguments."""
    p = argparse.ArgumentParser(description=__doc__, epilog=EPILOG)

    p.add_argument("db", type=str, help="name of database to phamerate")

    s = p.add_argument_group("MMseqs2 sequence-sequence clustering arguments")
    s.add_argument("--identity",
                   type=float, default=SEQ_MIN_SEQ_ID, metavar='',
                   help=f"percent identity for sequence-sequence clustering "
                        f"[default: {SEQ_MIN_SEQ_ID}%%]")
    s.add_argument("--coverage",
                   type=float, default=SEQ_COVERAGE, metavar='',
                   help=f"percent coverage for sequence-sequence clustering "
                        f"[default: {SEQ_COVERAGE}%%]")
    s.add_argument("--evalue",
                   type=float, default=SEQ_EVALUE, metavar='',
                   help=f"E-value threshold for sequence-sequence clustering "
                        f"[default: {SEQ_EVALUE}]")
    s.add_argument("--sensitivity",
                   type=float, default=SEQ_SENSITIVITY, metavar='',
                   help=f"sensitivity: 1 favors speed, 7 favors "
                        f"sensitivity [default: {SEQ_SENSITIVITY}]")
    s.add_argument("--cluster-mode",
                   type=int, default=SEQ_CLUSTER_MODE, metavar='',
                   help=f"clustering algorithm [default: {SEQ_CLUSTER_MODE}]")
    s.add_argument("--cluster-steps",
                   type=int, default=SEQ_CLUSTER_STEPS, metavar='',
                   help=f"number of steps for sequence-sequence clustering "
                        f"to proceed in [default: {SEQ_CLUSTER_STEPS}]")

    h = p.add_argument_group("MMseqs2 profile-sequence clustering arguments")
    h.add_argument("--hmm-identity",
                   type=float, default=HMM_MIN_SEQ_ID, metavar='',
                   help=f"percent identity for profile-consensus clustering "
                        f"[default: {HMM_MIN_SEQ_ID}%%]")
    h.add_argument("--hmm-coverage",
                   type=float, default=HMM_COVERAGE, metavar='',
                   help=f"percent coverage for profile-consensus clustering "
                        f"[default: {HMM_COVERAGE}%%]")
    h.add_argument("--hmm-evalue",
                   type=float, default=HMM_EVALUE, metavar='',
                   help=f"E-value threshold for profile-consensus clustering "
                        f"[default: {HMM_EVALUE}]")
    h.add_argument("--hmm-sensitivity",
                   type=float, default=HMM_SENSITIVITY, metavar='',
                   help=f"sensitivity: 1 favors speed, 7 favors "
                        f"sensitivity [default: {HMM_SENSITIVITY}]")
    h.add_argument("--hmm-cluster-mode",
                   type=int, default=HMM_CLUSTER_MODE, metavar='',
                   help=f"clustering algorithm [default: {HMM_CLUSTER_MODE}]")
    h.add_argument("--hmm-cluster-steps",
                   type=int, default=HMM_CLUSTER_STEPS, metavar='',
                   help=f"number of steps for profile-consensus clustering "
                        f"to proceed in [default: {HMM_CLUSTER_STEPS}]")
    h.add_argument("--skip-hmm", action="store_true",
                   help="do not perform profile-consensus clustering")

    p.add_argument("-v", "--verbose", action="store_true",
                   help="print progress messages to the console")
    p.add_argument("-d", "--debug", action="store_true",
                   help="run in debug mode")
    p.add_argument("-x", "--reset", action="store_true",
                   help="reset existing pham data before running")
    p.add_argument("-t", "--threads", type=int, default=CPUS,
                   help=f"number of threads to use [default: {CPUS}]")
    p.add_argument("-c", "--config-file", type=pathlib.Path, default=None,
                   help="path to file containing MySQL login details")

    return p


def setup_mmseqs_params(args):
    """Parse MMseqs2-related arguments into dictionaries of clustering
    parameters.

    :param args: dictionary of commandline arguments
    :type args: dict[str, int or float]
    :return: seq_params, hmm_params
    """
    seq_params = {"identity":         args["identity"],
                  "coverage":         args["coverage"],
                  "evalue":           args["evalue"],
                  "sensitivity":      args["sensitivity"],
                  "cluster_mode":     args["cluster_mode"],
                  "cluster_steps":    args["cluster_steps"]}

    if args["skip_hmm"]:
        hmm_params = None
    else:
        hmm_params = {"identity":      args["hmm_identity"],
                      "coverage":      args["hmm_coverage"],
                      "evalue":        args["hmm_evalue"],
                      "sensitivity":   args["sensitivity"],
                      "cluster_mode":  args["hmm_cluster_mode"],
                      "cluster_steps": args["hmm_cluster_steps"]}

    return seq_params, hmm_params


def main(argument_list):
    """Commandline entrypoint to this module."""
    # Set up the argument parser
    parser = setup_argparser()

    # Parse arguments
    args = vars(parser.parse_args(argument_list[2:]))

    # Create config object with data obtained from file and/or defaults.
    config = build_complete_config(args["config_file"])
    mysql_creds = config["mysql"]

    # Parse MMseqs2 parameters
    seq_params, hmm_params = setup_mmseqs_params(args)

    # Other runtime params
    verbose = args["verbose"]
    debug = args["debug"]
    threads = args["threads"]
    reset = args["reset"]

    # Record start time
    start_time = datetime.now()

    # Initialize SQLAlchemy engine with database provided at CLI
    alchemist = AlchemyHandler(database=args["db"],
                               username=mysql_creds["user"],
                               password=mysql_creds["password"])
    alchemist.connect(login_attempts=5, pipeline=True)
    engine = alchemist.engine

    # If reset, execute pham reset statements
    if reset:
        engine.execute("UPDATE gene SET PhamID = NULL")
        engine.execute("DELETE FROM pham")

    # Get old pham data and un-phamerated genes
    old_phams = get_pham_geneids(engine)
    old_colors = get_pham_colors(engine)
    new_genes = get_new_geneids(engine)

    # Create a SequenceDB
    database = SequenceDB()
    for geneid, translation in get_geneids_and_translations(engine).items():
        database.add_gene(geneid, translation)

    # Print initial snapshot of MySQL database
    if verbose:
        initial_summary = \
            DB_SUMMARY.format(len(old_phams),
                              sum([len(x) == 1 for x in old_phams.values()]),
                              sum([len(x) for x in old_phams.values()]),
                              len(new_genes), len(database),
                              len(database.translations))
        print(initial_summary)

    phams = assemble_phams(database, seq_params, hmm_params,
                           cpus=threads, verbose=verbose, debug=debug)

    new_phams = dict()
    for i, pham in enumerate(phams):
        pham_geneids = set(pham.geneids)
        new_phams[i+1] = pham_geneids

    # Preserve old pham names and colors
    print("Preserving old phamily names/colors where possible...")
    new_phams, new_colors = preserve_phams(old_phams, new_phams,
                                           old_colors, new_genes)

    # Early exit if we don't have new phams or new colors - avoids
    # overwriting the existing pham data with incomplete new data
    if len(new_phams) == 0 or len(new_colors) == 0:
        print("Failed to parse new pham/color data properly... Terminating "
              "pipeline")
        return

    # Update gene/pham tables with new pham data. Pham colors need to be done
    # first, because gene.PhamID is a foreign key to pham.PhamID.
    print("Updating pham data in database...")
    update_pham_table(new_colors, engine)
    update_gene_table(new_phams, engine)

    # Fix miscolored phams/orphams
    print("Phixing phalsely phlagged orphams...", end=" ")
    fix_white_phams(engine)
    print("Phixing phalsely hued phams...", end=" ")
    fix_colored_orphams(engine)

    # Close all connections in the connection pool.
    engine.dispose()

    # Print final snapshot of MySQL database
    if verbose:
        final_summary = \
            DB_SUMMARY.format(len(phams),
                              sum([len(x) == 1 for x in phams]),
                              sum([len(x) for x in phams]), 0,
                              len(database), len(database.translations))
        print(final_summary)

    # Record stop time
    stop_time = datetime.now()
    elapsed_time = str(stop_time - start_time)

    # Report phameration elapsed time
    print(f"Elapsed time: {elapsed_time}")


if __name__ == "__main__":
    main(sys.argv[1:])



"""Empty database schema DDL for schema version 0."""

STATEMENTS = [
    """/*!40101 SET @OLD_CHARACTER_SET_CLIENT=@@CHARACTER_SET_CLIENT */;""",
    """/*!40101 SET @OLD_CHARACTER_SET_RESULTS=@@CHARACTER_SET_RESULTS */;""",
    """/*!40101 SET @OLD_COLLATION_CONNECTION=@@COLLATION_CONNECTION */;""",
    """/*!40101 SET NAMES utf8 */;""",
    """/*!40103 SET @OLD_TIME_ZONE=@@TIME_ZONE */;""",
    """/*!40103 SET TIME_ZONE='+00:00' */;""",
    """/*!40014 SET @OLD_UNIQUE_CHECKS=@@UNIQUE_CHECKS, UNIQUE_CHECKS=0 */;""",
    """/*!40014 SET @OLD_FOREIGN_KEY_CHECKS=@@FOREIGN_KEY_CHECKS, FOREIGN_KEY_CHECKS=0 */;""",
    """/*!40101 SET @OLD_SQL_MODE=@@SQL_MODE, SQL_MODE='NO_AUTO_VALUE_ON_ZERO' */;""",
    """/*!40111 SET @OLD_SQL_NOTES=@@SQL_NOTES, SQL_NOTES=0 */;""",
    """DROP TABLE IF EXISTS `domain`;""",
    """/*!40101 SET @saved_cs_client     = @@character_set_client */;""",
    """/*!40101 SET character_set_client = utf8 */;""",
    """CREATE TABLE `domain` (
          `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
          `hit_id` varchar(25) NOT NULL,
          `description` blob,
          `DomainID` varchar(10) DEFAULT NULL,
          `Name` varchar(25) DEFAULT NULL,
          PRIMARY KEY (`id`),
          UNIQUE KEY `hit_id` (`hit_id`)
        ) ENGINE=InnoDB AUTO_INCREMENT=2186092 DEFAULT CHARSET=latin1;""",
    """/*!40101 SET character_set_client = @saved_cs_client */;""",
    """DROP TABLE IF EXISTS `gene`;""",
    """/*!40101 SET @saved_cs_client     = @@character_set_client */;""",
    """/*!40101 SET character_set_client = utf8 */;""",
    """CREATE TABLE `gene` (
          `GeneID` varchar(35) NOT NULL DEFAULT '',
          `PhageID` varchar(25) NOT NULL,
          `Start` mediumint(9) NOT NULL,
          `Stop` mediumint(9) NOT NULL,
          `Length` mediumint(9) NOT NULL,
          `Name` varchar(50) NOT NULL,
          `TypeID` varchar(10) DEFAULT NULL,
          `translation` varchar(4000) DEFAULT NULL,
          `StartCodon` enum('ATG','GTG','TTG') DEFAULT NULL,
          `StopCodon` enum('TAA','TAG','TGA') DEFAULT NULL,
          `Orientation` enum('F','R') DEFAULT NULL,
          `GC1` float DEFAULT NULL,
          `GC2` float DEFAULT NULL,
          `GC3` float DEFAULT NULL,
          `GC` float DEFAULT NULL,
          `LeftNeighbor` varchar(25) DEFAULT NULL,
          `RightNeighbor` varchar(25) DEFAULT NULL,
          `Notes` blob,
          `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
          `clustalw_status` enum('avail','pending','stale','done') NOT NULL DEFAULT 'avail',
          `blast_status` enum('avail','pending','stale','done') NOT NULL DEFAULT 'avail',
          PRIMARY KEY (`GeneID`),
          KEY `PhageID` (`PhageID`),
          KEY `id` (`id`),
          CONSTRAINT `gene_ibfk_1` FOREIGN KEY (`PhageID`) REFERENCES `phage` (`PhageID`)
        ) ENGINE=InnoDB AUTO_INCREMENT=1148013 DEFAULT CHARSET=latin1;""",
    """/*!40101 SET character_set_client = @saved_cs_client */;""",
    """DROP TABLE IF EXISTS `gene_domain`;""",
    """/*!40101 SET @saved_cs_client     = @@character_set_client */;""",
    """/*!40101 SET character_set_client = utf8 */;""",
    """CREATE TABLE `gene_domain` (
          `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
          `GeneID` varchar(35) DEFAULT NULL,
          `hit_id` varchar(25) NOT NULL,
          `query_start` int(10) unsigned NOT NULL,
          `query_end` int(10) unsigned NOT NULL,
          `expect` double unsigned NOT NULL,
          PRIMARY KEY (`id`),
          UNIQUE KEY `GeneID__hit_id` (`GeneID`,`hit_id`),
          KEY `hit_id` (`hit_id`),
          CONSTRAINT `gene_domain_ibfk_1` FOREIGN KEY (`GeneID`) REFERENCES `gene` (`GeneID`),
          CONSTRAINT `gene_domain_ibfk_2` FOREIGN KEY (`hit_id`) REFERENCES `domain` (`hit_id`)
        ) ENGINE=InnoDB AUTO_INCREMENT=1395804 DEFAULT CHARSET=latin1;""",
    """/*!40101 SET character_set_client = @saved_cs_client */;""",
    """DROP TABLE IF EXISTS `host`;""",
    """/*!40101 SET @saved_cs_client     = @@character_set_client */;""",
    """/*!40101 SET character_set_client = utf8 */;""",
    """CREATE TABLE `host` (
          `ID` int(11) NOT NULL AUTO_INCREMENT,
          `Name` varchar(50) NOT NULL,
          `Accession` varchar(15) DEFAULT NULL,
          PRIMARY KEY (`ID`)
        ) ENGINE=InnoDB DEFAULT CHARSET=latin1;""",
    """/*!40101 SET character_set_client = @saved_cs_client */;""",
    """DROP TABLE IF EXISTS `host_range`;""",
    """/*!40101 SET @saved_cs_client     = @@character_set_client */;""",
    """/*!40101 SET character_set_client = utf8 */;""",
    """CREATE TABLE `host_range` (
          `PhageID` varchar(25) NOT NULL,
          `host_id` int(11) NOT NULL,
          `plating_efficiency` double unsigned NOT NULL,
          `below_detection` tinyint(1) NOT NULL,
          KEY `PhageID` (`PhageID`),
          KEY `host_id` (`host_id`),
          CONSTRAINT `host_range_ibfk_1` FOREIGN KEY (`PhageID`) REFERENCES `phage` (`PhageID`),
          CONSTRAINT `host_range_ibfk_2` FOREIGN KEY (`host_id`) REFERENCES `host` (`ID`)
        ) ENGINE=InnoDB DEFAULT CHARSET=latin1;""",
    """/*!40101 SET character_set_client = @saved_cs_client */;""",
    """DROP TABLE IF EXISTS `node`;""",
    """/*!40101 SET @saved_cs_client     = @@character_set_client */;""",
    """/*!40101 SET character_set_client = utf8 */;""",
    """CREATE TABLE `node` (
          `id` int(11) unsigned NOT NULL AUTO_INCREMENT,
          `platform` varchar(15) DEFAULT NULL,
          `hostname` varchar(50) DEFAULT NULL,
          PRIMARY KEY (`id`),
          KEY `hostname` (`hostname`)
        ) ENGINE=InnoDB DEFAULT CHARSET=latin1;""",
    """/*!40101 SET character_set_client = @saved_cs_client */;""",
    """DROP TABLE IF EXISTS `phage`;""",
    """/*!40101 SET @saved_cs_client     = @@character_set_client */;""",
    """/*!40101 SET character_set_client = utf8 */;""",
    """CREATE TABLE `phage` (
          `PhageID` varchar(25) NOT NULL,
          `Accession` varchar(15) NOT NULL,
          `Name` varchar(50) NOT NULL,
          `Isolated` varchar(100) DEFAULT NULL,
          `HostStrain` varchar(50) DEFAULT NULL,
          `Sequence` mediumblob NOT NULL,
          `SequenceLength` mediumint(9) NOT NULL,
          `Prophage` enum('yes','no') DEFAULT NULL,
          `ProphageOffset` int(11) DEFAULT NULL,
          `DateLastModified` datetime DEFAULT NULL,
          `DateLastSearched` datetime DEFAULT NULL,
          `Notes` blob,
          `GC` float DEFAULT NULL,
          `Cluster` varchar(5) DEFAULT NULL,
          `status` varchar(5) DEFAULT NULL,
          PRIMARY KEY (`PhageID`)
        ) ENGINE=InnoDB DEFAULT CHARSET=latin1;""",
    """/*!40101 SET character_set_client = @saved_cs_client */;""",
    """DROP TABLE IF EXISTS `pham`;""",
    """/*!40101 SET @saved_cs_client     = @@character_set_client */;""",
    """/*!40101 SET character_set_client = utf8 */;""",
    """CREATE TABLE `pham` (
          `GeneID` varchar(35) NOT NULL DEFAULT '',
          `name` int(10) unsigned DEFAULT NULL,
          `orderAdded` int(5) unsigned NOT NULL AUTO_INCREMENT,
          PRIMARY KEY (`GeneID`),
          KEY `name_index` (`name`),
          KEY `orderAdded` (`orderAdded`),
          CONSTRAINT `pham_ibfk_1` FOREIGN KEY (`GeneID`) REFERENCES `gene` (`GeneID`)
        ) ENGINE=InnoDB DEFAULT CHARSET=latin1;""",
    """/*!40101 SET character_set_client = @saved_cs_client */;""",
    """DROP TABLE IF EXISTS `pham_color`;""",
    """/*!40101 SET @saved_cs_client     = @@character_set_client */;""",
    """/*!40101 SET character_set_client = utf8 */;""",
    """CREATE TABLE `pham_color` (
          `id` int(10) unsigned NOT NULL AUTO_INCREMENT,
          `name` int(10) unsigned NOT NULL,
          `color` char(7) NOT NULL,
          PRIMARY KEY (`id`)
        ) ENGINE=InnoDB AUTO_INCREMENT=23752 DEFAULT CHARSET=latin1;""",
    """/*!40101 SET character_set_client = @saved_cs_client */;""",
    """DROP TABLE IF EXISTS `pham_history`;""",
    """/*!40101 SET @saved_cs_client     = @@character_set_client */;""",
    """/*!40101 SET character_set_client = utf8 */;""",
    """CREATE TABLE `pham_history` (
          `name` int(10) unsigned NOT NULL,
          `parent` int(10) unsigned NOT NULL,
          `action` enum('join','split') NOT NULL,
          `datetime` timestamp NOT NULL DEFAULT CURRENT_TIMESTAMP,
          PRIMARY KEY (`name`,`parent`,`action`),
          KEY `parent` (`parent`)
        ) ENGINE=InnoDB DEFAULT CHARSET=latin1;""",
    """/*!40101 SET character_set_client = @saved_cs_client */;""",
    """DROP TABLE IF EXISTS `pham_old`;""",
    """/*!40101 SET @saved_cs_client     = @@character_set_client */;""",
    """/*!40101 SET character_set_client = utf8 */;""",
    """CREATE TABLE `pham_old` (
          `GeneID` varchar(35) DEFAULT NULL,
          `name` int(10) unsigned DEFAULT NULL,
          `orderAdded` int(5) unsigned NOT NULL AUTO_INCREMENT,
          PRIMARY KEY (`orderAdded`),
          KEY `orderAdded` (`orderAdded`),
          KEY `GeneID` (`GeneID`),
          CONSTRAINT `pham_old_ibfk_1` FOREIGN KEY (`GeneID`) REFERENCES `gene` (`GeneID`) ON UPDATE CASCADE
        ) ENGINE=InnoDB DEFAULT CHARSET=latin1;""",
    """/*!40101 SET character_set_client = @saved_cs_client */;""",
    """DROP TABLE IF EXISTS `scores_summary`;""",
    """/*!40101 SET @saved_cs_client     = @@character_set_client */;""",
    """/*!40101 SET character_set_client = utf8 */;""",
    """CREATE TABLE `scores_summary` (
          `id` bigint(20) NOT NULL AUTO_INCREMENT,
          `query` varchar(35) DEFAULT NULL,
          `subject` varchar(35) DEFAULT NULL,
          `blast_score` double unsigned DEFAULT NULL,
          `clustalw_score` decimal(5,4) unsigned DEFAULT NULL,
          `blast_bit_score` double unsigned DEFAULT NULL,
          PRIMARY KEY (`id`),
          KEY `query` (`query`),
          KEY `subject` (`subject`),
          CONSTRAINT `scores_summary_ibfk_1` FOREIGN KEY (`query`) REFERENCES `gene` (`GeneID`),
          CONSTRAINT `scores_summary_ibfk_2` FOREIGN KEY (`subject`) REFERENCES `gene` (`GeneID`)
        ) ENGINE=InnoDB DEFAULT CHARSET=latin1;""",
    """/*!40101 SET character_set_client = @saved_cs_client */;""",
    """/*!40103 SET TIME_ZONE=@OLD_TIME_ZONE */;""",
    """/*!40101 SET SQL_MODE=@OLD_SQL_MODE */;""",
    """/*!40014 SET FOREIGN_KEY_CHECKS=@OLD_FOREIGN_KEY_CHECKS */;""",
    """/*!40014 SET UNIQUE_CHECKS=@OLD_UNIQUE_CHECKS */;""",
    """/*!40101 SET CHARACTER_SET_CLIENT=@OLD_CHARACTER_SET_CLIENT */;""",
    """/*!40101 SET CHARACTER_SET_RESULTS=@OLD_CHARACTER_SET_RESULTS */;""",
    """/*!40101 SET COLLATION_CONNECTION=@OLD_COLLATION_CONNECTION */;""",
    """/*!40111 SET SQL_NOTES=@OLD_SQL_NOTES */;"""
    ]



"""Misc. base/simple functions. These should not require import of other
modules in this package to prevent circular imports."""

from collections import OrderedDict
import csv
import getpass
import os
from pathlib import Path
import sys

from pdm_utils.constants import constants


def find_expression(expression, list_of_items):
    """Counts the number of items with matches to a regular expression.

    :param expression: Regular expression object
    :type expression: re
    :param list_of_items:
        List of items that will be searched with the regular expression.
    :type list_of_items: list
    :returns:
        Number of times the regular expression was identified in the list.
    :rtype: int
    """
    search_tally = 0
    for element in list_of_items:
        search_result = expression.search(element)
        if search_result:
            search_tally += 1
    return search_tally


def edit_suffix(value, option, suffix=constants.NAME_SUFFIX):
    """Adds or removes the indicated suffix to an input value.

    :param value: Value that will be edited.
    :type value: str
    :param option:
        Indicates what to do with the value and suffix ('add', 'remove').
    :type option: str
    :param suffix: The suffix that will be added or removed.
    :type suffix: str
    :returns:
        The edited value. The suffix is not added if the input
        value already has the suffix.
    :rtype: str
    """

    if option.lower() == "add":
        if not value.lower().endswith(suffix.lower()):
            value = value + suffix
    elif option.lower() == "remove":
        if value.lower().endswith(suffix.lower()):
            value = value[:-len(suffix)]
    else:
        pass
    return value


def create_indices(input_list, batch_size):
    """Create list of start and stop indices to split a list into batches.

    :param input_list: List from which to generate batch indices.
    :type input_list: list
    :param batch_size: Size of each batch.
    :type batch_size: int
    :returns: List of 2-element tuples (start index, stop index).
    :rtype: list
    """
    index_list = []
    for start in range(0, len(input_list), batch_size):
        if start + batch_size > len(input_list):
            stop = len(input_list)
        else:
            stop = start + batch_size
        tup = (start, stop)
        index_list.append(tup)
    return index_list


def reformat_strand(input_value, format, case=False):
    """Converts common strand orientation formats.


    :param input_value: Value that will be edited.
    :type input_value: str, int
    :param format:
        Indicates how the value should be edited.
        Valid format types include:
        'fr_long' ('forward', 'reverse')
        'fr_short' ('f', 'r')
        'fr_abbrev1' ('for', 'rev')
        'fr_abbrev2' ('fwd', 'rev')
        'tb_long' ('top', 'bottom')
        'tb_short' ('t', 'b')
        'wc_long' ('watson', 'crick')
        'wc_short' ('w','c')
        'operator' ('+', '-')
        'numeric' (1, -1).
    :type format: str
    :param case:
        Indicates whether the output value should be capitalized.
    :type case: bool
    :returns: The re-formatted value as indicated by 'format'.
    :rtype: str, int
    """

    format_dict = {"fr_long":["forward", "reverse"],
                    "fr_short":["f", "r"],
                    "fr_abbrev1":["for", "rev"],
                    "fr_abbrev2":["fwd", "rev"],
                    "tb_long":["top", "bottom"],
                    "tb_short":["t", "b"],
                    "wc_long":["watson", "crick"],
                    "wc_short":["w", "c"],
                    "operator":["+", "-"],
                    "numeric":[1, -1]}
    strand1_values = set()
    strand2_values = set()
    for values in format_dict.values():
        strand1_values.add(values[0])
        strand2_values.add(values[1])
    if format in format_dict.keys():
        if isinstance(input_value, str):
            new_value = input_value.lower()
        else:
            new_value = input_value
        if new_value in strand1_values:
            new_value = format_dict[format][0]
        elif new_value in strand2_values:
            new_value = format_dict[format][1]
        else:
            new_value = "NA"
    else:
        new_value = "NA"
    if isinstance(new_value, str):
        if case == True:
            new_value = new_value.capitalize()
    return new_value


def reformat_coordinates(start, stop, current, new):
    """Converts common coordinate formats.

    The type of coordinate formats include:

    '0_half_open':

        0-based half-open intervals that is the common format
        for BAM files and UCSC Browser database. This format seems
        to be more efficient when performing genomics
        computations.

    '1_closed':

        1-based closed intervals that is the common format
        for the MySQL Database, UCSC Browser,
        the Ensembl genomics database,
        VCF files, GFF files. This format seems to be more
        intuitive and used for visualization.

    The function assumes coordinates reflect the start and stop
    boundaries (where the start coordinates is smaller than the stop
    coordinate), instead of transcription start and stop coordinates.

    :param start: Start coordinate
    :type start: int
    :param stop: Stop coordinate
    :type stop: int
    :param current: Indicates the indexing format of the input coordinates.
    :type current: str
    :param new: Indicates the indexing format of the output coordinates.
    :type new: str
    :returns: The re-formatted start and stop coordinates.
    :rtype: int
    """
    format_set = set(["0_half_open", "1_closed"])
    if (current in format_set and new in format_set):
        if current == "0_half_open":
            if new == "1_closed":
                new_start = start + 1
                new_stop = stop
            else:
                new_start = start
                new_stop = stop
        else:
            if new == "0_half_open":
                new_start = start - 1
                new_stop = stop
            else:
                new_start = start
                new_stop = stop
    else:
        # new_start = start
        # new_stop = stop
        raise ValueError("Format for CDS coordinate formats must "
                         "either be '0_half_open' or '1_closed'")
    return (new_start, new_stop)


# TODO this may no longer be needed.
def check_empty(value, lower=True):
    """Checks if the value represents a null value.

    :param value: Value to be checked against the empty set.
    :type value: misc.
    :param lower:
        Indicates whether the input value should be lowercased
        prior to checking.
    :type lower: bool
    :returns: Indicates whether the value is present in the empty set.
    :rtype: bool
    """
    if (lower == True and isinstance(value, str)):
        value = value.lower()
    if value in constants.EMPTY_SET:
        result = True
    else:
        result = False
    return result


def convert_empty(input_value, format, upper=False):
    """Converts common null value formats.


    :param input_value: Value to be re-formatted.
    :type input_value: str, int, datetime
    :param format:
        Indicates how the value should be edited.
        Valid format types include:
        'empty_string' = ''
        'none_string' = 'none'
        'null_string' = 'null'
        'none_object' = None
        'na_long' = 'not applicable'
        'na_short' = 'na'
        'n/a' = 'n/a'
        'zero_string' = '0'
        'zero_num' = 0
        'empty_datetime_obj' = datetime object with arbitrary date, '1/1/0001'
    :type format: str
    :param upper:
        Indicates whether the output value should be uppercased.
    :type upper: bool
    :returns: The re-formatted value as indicated by 'format'.
    :rtype: str, int, datetime
    """
    format_dict = {"empty_string": "",
                    "none_string": "none",
                    "null_string": "null",
                    "none_object": None,
                    "na_long": "not applicable",
                    "na_short": "na",
                    "n/a": "n/a",
                    "zero_string": "0",
                    "zero_num": 0,
                    "empty_datetime_obj": constants.EMPTY_DATE}
    output_values = set(format_dict.values())
    if format in format_dict.keys():

        # Convert input value to ensure it matches the formatting of
        # all possible output values.
        if isinstance(input_value, str):
            new_value = input_value.lower()
        else:
            new_value = input_value

        # Now check to see if the value is present in among all possible
        # output values, and convert.
        if new_value in output_values:
            new_value = format_dict[format]
        else:
            new_value = input_value
    else:
        new_value = input_value
    if isinstance(new_value, str):
        if upper == True:
            new_value = new_value.upper()
    return new_value


def reformat_description(raw_description):
    """Reformat a gene description.

    :param raw_description: Input value to be reformatted.
    :type raw_description: str
    :returns:
        tuple (description, processed_description)
        WHERE
        description(str) is the original value stripped of leading
        and trailing whitespace.
        processed_description(str) is the reformatted value, in which
        non-informative/generic data is removed.
    :rtype: tuple
    """
    if raw_description is None:
        raw_description = ""
    raw_description = raw_description.strip()
    processed_description = raw_description.lower()
    split_description = processed_description.split(" ")
    generic_set = {
        "conserved hypothetical protein",
        "conserved phage membrane protein",
        "conserved phage protein",
        "conserved protein",
        "gp",
        "hypothetical",
        "hypothetical phage protein",
        "hypothetical phageprotein",
        "hypothetical phage membrane protein",
        "hypothetical protein",
        "integral membrane protein",
        "membrane protein",
        "orf",
        "phage membrane protein",
        "phage protein",
        "protein",
        "putative",
        "regulatory protein",
        "structural protein",
        "unknown",
        "\\n",
        "."
        }

    if processed_description in generic_set:
        processed_description = ""
    elif is_float(processed_description):
        processed_description = ""
    elif processed_description.startswith("original glimmer"):
        processed_description = ""
    elif processed_description.startswith("original genemark"):
        processed_description = ""
    elif len(split_description) == 1:

        # Check if similar to gp10, ORF10, ORF10.1
        first_left, first_right = split_string(split_description[0])
        if (first_left in generic_set and first_right != ""):
            processed_description = ""
        else:
            processed_description = raw_description

    elif len(split_description) == 2:

        first_left, first_right = split_string(split_description[0])
        second_left, second_right = split_string(split_description[1])

        # gp 10
        if (first_left in generic_set and second_right != ""):
            processed_description = ""
        # putative protein, hypothetical protein, putative gp10)
        elif (first_left in generic_set and second_left in generic_set):
            processed_description = ""
        else:
            processed_description = raw_description
    else:
        processed_description = raw_description
    return (raw_description, processed_description)


def identify_unique_items(complete_list):
    """Identify unique and non-unique items in a list.

    :param complete_list: List of items that will be evaluated.
    :type complete_list: list
    :returns:
        tuple (unique_set, duplicate_set)
        WHERE
        unique_set(set) is a set of all unique/non-duplicated items.
        duplicate_set(set) is a set of non-unique/duplicated items.
        non-informative/generic data is removed.
    :rtype: tuple
    """
    unique_set = set() # Set of all non-duplicated unique items.
    duplicate_set = set() # Set of all duplicated items.
    for item in complete_list:
        if item not in unique_set:
            unique_set.add(item)
        else:
            duplicate_set.add(item)
    # Remove items from the unique list that eventually were duplicated.
    unique_set = unique_set - duplicate_set
    return (unique_set, duplicate_set)


# TODO implement? this may no longer be needed.
def identify_nested_items(complete_list):
    """Identify nested and non-nested two-element tuples in a list.

    :param complete_list: List of tuples that will be evaluated.
    :type complete_list: list
    :returns:
        tuple (not_nested_set, nested_set)
        WHERE
        not_nested_set(set) is a set of non-nested tuples.
        nested_set(set) is a set of nested tuples.
    :rtype: tuple
    """
    not_nested_set = set()
    nested_set = set()
    return (not_nested_set, nested_set)


def trim_characters(string):
    """Remove leading and trailing generic characters from a string.

    :param string:
        Value that will be trimmed.
        Characters that will be removed include: '.', ',', ';', '-', '_'.
    :type string: str
    :returns: Edited value.
    :rtype: str
    """
    generic_characters = set([".", ",", ";", "-", "_"])
    if len(string) > 0:
        # Trim non-specific trailing characters.
        value1 = True
        while value1:
            if string[-1] in generic_characters:
                string = string[:-1]
            else:
                value1 = False
        # Trim non-specific leading characters.
        value2 = True
        while value2:
            if string[0] in generic_characters:
                string = string[1:]
            else:
                value2 = False
    return string


def parse_names_from_record_field(description):
    """
    Attempts to parse the phage/plasmid/prophage name and host genus
    from a given string.
    :param description: the input string to be parsed
    :type description: str
    :return: name, host_genus
    """
    generic_words = ("phage", "unclassified", "complete", "genome", "sequence")
    host_genus = ""
    name = ""

    # Split description into list, and trim leading and trailing generic chars
    description = [trim_characters(x) for x in description.strip().split()]

    # If there's only a non-generic word in the description, take that as name
    if len(description) == 1 and description[0].lower() not in generic_words:
        if description[0].lower().endswith("phage"):
            host_genus = description[0][:-5]
        else:
            name = description[0]

    # If there are two words in the description, look for two options:
    elif len(description) == 2:
        # Parses format: Mycobacteriophage Trixie or phage Trixie or virus Trixie
        if description[0].lower().endswith('phage') or \
                description[0].lower().endswith('virus'):
            host_genus = description[0][:-5]
            name = description[1]
        # Parses format: Trixie_Draft Unclassified
        elif description[0].lower() not in generic_words and \
                description[1].lower() == "unclassified":
            name = description[0]
        # Parses format: Mycobacterium phage
        elif description[1].lower() in ("phage", "prophage", "plasmid", "virus"):
            host_genus = description[0]
    # If there are multiple words in the description, look for replicon type
    else:
        for i, value in enumerate(description):
            # Parses format: Mycobacterium phage Trixie
            if value.lower() in ("phage", "prophage", "plasmid", "virus"):
                if i > 0:
                    host_genus = description[i-1]
                if i + 1 < len(description):
                    name = description[i+1]

    return name, host_genus


def compare_sets(set1, set2):
    """Compute the intersection and differences between two sets.

    :param set1: The first input set.
    :type set1: set
    :param set2: The second input set.
    :type set2: set
    :returns:
        tuple (set_intersection, set1_diff, set2_diff)
        WHERE
        set_intersection(set) is the set of shared values.
        set1_diff(set) is the set of values unique to the first set.
        set2_diff(set) is the set of values unique to the second set.
    :rtype: tuple
    """
    set_intersection = set1 & set2
    set1_diff = set1 - set2
    set2_diff = set2 - set1
    return set_intersection, set1_diff, set2_diff


def match_items(list1, list2):
    """Match values of two lists and return several results.

    :param list1: The first input list.
    :type list1: list
    :param list2: The second input list.
    :type list2: list
    :returns:
        tuple (matched_unique_items, set1_unmatched_unique_items,
        set2_unmatched_unique_items, set1_duplicate_items,
        set2_duplicate_items)
        WHERE
        matched_unique_items(set) is the set of matched unique values.
        set1_unmatched_unique_items(set) is the set of
        unmatched unique values from the first list.
        set2_unmatched_unique_items(set) is the set of
        unmatched unique values from the second list.
        set1_duplicate_items(set) is the the set of
        duplicate values from the first list.
        set2_duplicate_items(set) is the set of
        unmatched unique values from the second list.
    :rtype: tuple
    """

    # Determine the unique values in each list.
    set1_unique_items, set1_duplicate_items = \
        identify_unique_items(list1)
    set2_unique_items, set2_duplicate_items = \
        identify_unique_items(list2)

    # Create matched and difference sets.
    matched_unique_items, \
    set1_unmatched_unique_items, \
    set2_unmatched_unique_items = \
        compare_sets(set1_unique_items, set2_unique_items)
    return (matched_unique_items,
            set1_unmatched_unique_items,
            set2_unmatched_unique_items,
            set1_duplicate_items,
            set2_duplicate_items)


def is_float(string):
    """Check if string can be converted to float."""
    try:
        float(string)
        return True
    except ValueError:
        return False


def split_string(string):
    """Split a string based on alphanumeric characters.

    Iterates through a string, identifies the first position
    in which the character is a float, and creates two strings at this
    position.

    :param string: The value to be split.
    :type string: str
    :returns:
        tuple (left, right)
        WHERE
        left(str) is the left portion of the input value prior to the
        first numeric character and only contains alphabetic characters
        (or will be '').
        right(str) is the right portion of the input value after the
        first numeric character and only contains numeric characters
        (or will be '').
    :rtype: tuple
    """
    left = ""
    right = ""
    if string.isalpha():
        left = string
    elif is_float(string):
        right = string
    else:
        x = 0
        value = False
        while (value == False and x < len(string)):
            if (string[:x].isalpha() and is_float(string[x:])):
                left = string[:x]
                right = string[x:]
                value = True
            x += 1
    return left, right


# TODO this function is specific to genome-level data, so should
# it be move as a Genome class method?
def compare_cluster_subcluster(cluster, subcluster):
    """Check if a cluster and subcluster designation are compatible.

    :param cluster:
        The cluster value to be compared.
        'Singleton' and 'UNK' are lowercased.
    :type cluster: str
    :param subcluster: The subcluster value to be compared.
    :type subcluster: str
    :returns:
        The result of the evaluation, indicating whether the two
        values are compatible.
    :rtype: bool
    """
    result = True

    # If Singleton or Unknown Cluster, there should be no Subcluster.
    # Subcluster = '' is considered an error since it is the default
    # attribute value when the Genome class is instantiated.
    if cluster.capitalize() == "Singleton" or cluster == "UNK":
        if subcluster != "none":
            result = False

    # If not Singleton or Unknown, then if
    # there is a Subcluster designation, the Cluster should be part
    # of Subcluster data and the remainder should be a digit
    elif subcluster != "none":
        left, right = split_string(subcluster)
        if left != cluster or right.isdigit() == False:
            result = False
    else:
        pass
    return result


def identify_one_list_duplicates(item_list):
    """Identify duplicate items within a list.

    :param item_list: The input list to be checked.
    :type item_list: list
    :returns: The set of non-unique/duplicated items.
    :rtype: set
    """
    duplicate_items = set()
    item_set = set(item_list)
    for item1 in item_set:
        count = 0
        for item2 in item_list:
            if item1 == item2:
                count += 1
        if count > 1:
            duplicate_items.add(item1)
    return duplicate_items


def identify_two_list_duplicates(item1_list, item2_list):
    """Identify duplicate items between two lists.

    :param item1_list: The first input list to be checked.
    :type item1_list: list
    :param item2_list: The second input list to be checked.
    :type item2_list: list
    :returns:
        The set of non-unique/duplicated items between the two lists
        (but not duplicate items within each list).
    :rtype: set
    """
    item1_set = set(item1_list)
    item2_set = set(item2_list)
    item3_set = item1_set & item2_set
    return item3_set


def check_value_expected_in_set(value, set1, expect=True):
    """Check if a value is present within a set and if it is expected.

    :param value: The value to be checked.
    :type value: misc.
    :param set1: The reference set of values.
    :type set1: set
    :param expect: Indicates if 'value' is expected to be present in 'set1'.
    :type expect: bool
    :returns: The result of the evaluation.
    :rtype: bool
    """
    # Check if the value is present in the set.
    if value in set1:
        present = True
    else:
        present = False
    # Compare the presence/absence with what was expected.
    if expect and present:
        result = True
    elif not expect and not present:
        result = True
    else:
        result = False
    return result


def check_value_in_two_sets(value, set1, set2):
    """Check if a value is present within two sets.

    :param value: The value to be checked.
    :type value: misc.
    :param set1: The first reference set of values.
    :type set1: set
    :param set2: The second reference set of values.
    :type set2: set
    :returns:

        The result of the evaluation, indicating whether the
        value is present within:

            1. only the 'first' set
            2. only the 'second' set
            3. 'both' sets
            4. 'neither' set

    :rtype: str
    """
    present1 = False
    present2 = False
    if value in set1:
        present1 = True
    if value in set2:
        present2 = True
    if present1 and present2:
        result = "both"
    elif present1 and not present2:
        result = "first"
    elif not present1 and present2:
        result = "second"
    else:
        result = "neither"
    return result


def lower_case(value):
    """Return the value lowercased if it is within a specific set of values.

    :param value: The value to be checked.
    :type value: str
    :returns:
        The lowercased value if it is equivalent to
        'none', 'retrieve', or 'retain'.
    :rtype: str
    """
    lower_set = {"none", "retrieve", "retain"}
    if isinstance(value, str):
        if value.lower() in lower_set:
            value = value.lower()
    return value


def ask_yes_no(prompt="", response_attempt=1):
    """Function to get the user's yes/no response to a question.

    Accepts variations of yes/y, true/t, no/n, false/f, exit/quit/q.

    :param prompt: the question to ask the user.
    :type prompt: str
    :param response_attempt:

        The number of the number of attempts allowed before the
        function exits. This prevents the script from getting stuck in a loop.

    :type response_attempt: int
    :returns:

        The default is False (e.g. user hits Enter without typing
        anything else), but variations of yes or true responses will return
        True instead. If the response is 'exit' or 'quit', the loop is exited
        and None is returned.

    :rtype: bool, None
    """
    response_valid = False
    while response_valid is False and response_attempt > 0:
        response = input(prompt)
        response_attempt -= 1
        if response.lower() in ("yes", "y", "t", "true"):
            response = True
            response_valid = True
        elif response.lower() in ("no", "n", "f", "false", ""):
            response = False
            response_valid = True
        elif response.lower() in ("exit", "quit", "q"):
            response = None
            response_valid = True
        else:
            response = None
            print("Invalid response. Enter [yes/no/exit].")
    return response


def identify_contents(path_to_folder, kind=None, ignore_set=set()):
    """Create a list of filenames and/or folders from an indicated directory.

    :param path_to_folder: A valid directory path.
    :type path_to_folder: Path
    :param kind:

        ("file", "dir"), corresponding with paths to be
        checked as either files or directories.
    :type kind: str
    :param ignore_set:
        A set of strings representing file or folder names to ignore.
    :type ignore_set: set
    :returns: List of valid contents in the directory.
    :rtype: list
    """
    contents = []
    for item in path_to_folder.iterdir():
        item_path = Path(path_to_folder, item)
        if kind == "file":
            if item_path.is_file() and item.name not in ignore_set:
                contents.append(item)
        elif kind == "dir":
            if item_path.is_dir() and item.name not in ignore_set:
                contents.append(item)
        elif kind is None:
            if item.name not in ignore_set:
                contents.append(item)
        else:
            msg = (f"{kind} is not a valid kind (None, dir, file) "
                  "for this function.")
            print(msg)
            # Returns None so that it is clear it wasn't evaluated.
            contents = None
    return contents


def expand_path(input_path):
    """Convert a non-absolute path into an absolute path.

    :param input_path: The path to be expanded.
    :type input_path: str
    :returns: The expanded path.
    :rtype: str
    """
    # "~" needs to be expanded first
    home_dir = os.path.expanduser("~")
    if input_path[0] == "~":
        expanded_path = home_dir + input_path[1:]
    else:
        expanded_path = input_path
    # os.path.abspath will resolve ./ or ../ present in the path
    expanded_path = os.path.abspath(expanded_path)
    return expanded_path


def verify_path(filepath, kind=None):
    """Verifies that a given path exists.

    :param filepath: full path to the desired file/directory.
    :type filepath: str
    :param kind:

        ("file", "dir"), corresponding with paths to be
        checked as either files or directories.

    :type kind: str
    :return Boolean: True if path is verified, False otherwise.
    """
    if kind == "file":
        if os.path.isfile(filepath) is True:
            return True
        else:
            return False
    elif kind == "dir":
        if os.path.isdir(filepath) is True:
            return True
        else:
            return False
    elif kind is None:
        if os.path.exists(filepath) is True:
            return True
        else:
            return False
    else:
        print("{} is not a valid kind for this function.".format(kind),
              "Please try again",
              "use one of (None, dir, file).")
        return False


def verify_path2(path, kind=None, expect=True):
    """Verifies that a given path exists.

    :param path: path
    :type path: Path
    :param kind:

        ("file", "dir"), corresponding with paths to be
        checked as either files or directories.

    :type kind: str
    :param expect: Indicates if the path is expected to the indicated kind.
    :type expect: bool
    :returns:
        tuple (result, message)
        WHERE
        result(bool) indicates if the expectation was satisfied.
        message(str) is a description of the result.
    :rtype: tuple
    """
    if kind == "file":
        exists = path.is_file()
        if (exists and not expect):
            result = False
            msg = f"The file {path} already exists."
        elif (not exists and expect):
            result = False
            msg = f"The file {path} does not exist."
        else:
            result = True
            msg = None
    elif kind == "dir":
        exists = path.is_dir()
        if (exists and not expect):
            result = False
            msg = f"The directory {path} already exists."
        elif (not exists and expect):
            result = False
            msg = f"The directory {path} does not exist."
        else:
            result = True
            msg = None
    elif kind == None:
        exists = path.exists()
        if (exists and not expect):
            result = False
            msg = f"The {path} already exists."
        elif (not exists and expect):
            result = False
            msg = f"The {path} does not exist."
        else:
            result = True
            msg = None
    else:
        result = False
        msg = (f"{kind} is not a valid kind (None, dir, file) "
              "for this function.")

    return result, msg


def set_path(path, kind=None, expect=True):
    """Confirm validity of path argument.

    :param path: path
    :type path: Path
    :param kind:

        ("file", "dir"), corresponding with paths to be
        checked as either files or directories.

    :type kind: str
    :param expect: Indicates if the path is expected to the indicated kind.
    :type expect: bool
    :returns: Absolute path if valid, otherwise sys.exit is called.
    :rtype: Path
    """
    path = path.expanduser()
    path = path.resolve()
    result, msg = verify_path2(path, kind=kind, expect=expect)
    if not result:
        print(msg)
        sys.exit(1)
    else:
        return path


def make_new_dir(output_dir, new_dir, attempt=1, mkdir=True):
    """Make a new directory.

    Checks to verify the new directory name is valid and does not
    already exist. If it already exists, it attempts to extend
    the name with an integer suffix.

    :param output_dir:
        Full path to the directory where the new directory will be created.
    :type output_dir: Path
    :param new_dir: Name of the new directory to be created.
    :type new_dir: Path
    :param attempt: Number of attempts to create the directory.
    :type attempt: int
    :returns:
        If successful, the full path of the created directory.
        If unsuccessful, None.
    :rtype: Path, None
    """
    valid = False
    count = 0
    while (not valid and count < attempt):
        if count > 0:
            new_dir_mod = new_dir.stem + "_" + str(count)
            new_dir_mod = Path(new_dir_mod)
        else:
            new_dir_mod = new_dir
        new_path = Path(output_dir, new_dir_mod)
        if not new_path.is_dir():
            valid = True
            if mkdir:
                new_path.mkdir()
        count += 1
    if not valid:
        return None
    else:
        return new_path


# TODO this may not be needed, if a standard config file format is used.
def parse_flag_file(flag_file):
    """Parse a file to an evaluation flag dictionary.

    :param flag_file:
        A two-column csv-formatted file
        WHERE
        1. evaluation flag
        2. 'True' or 'False'
    :type flag_file: str
    :returns:
        A dictionary
        WHERE
        keys (str) are evaluation flags
        values (bool) indicate the flag setting
        Only flags that contain boolean values are returned.
    :rtype: dict
    """
    eval_flags = {}
    with open(flag_file,'r') as file:
        file_reader = csv.reader(file)
        for row in file_reader:
            if row[1].lower() == "true":
                row[1] = True
                eval_flags[row[0].lower()] = row[1]
            elif row[1].lower() == "false":
                row[1] = False
                eval_flags[row[0].lower()] = row[1]
            else:
                pass
    return eval_flags


def get_user_pwd(user_prompt="Username: ", pwd_prompt="Password: "):
    """Get username and password.

    :param user_prompt: Displayed description when prompted for username.
    :type user_prompt: str
    :param pwd_prompt: Displayed description when prompted for password.
    :type pwd_prompt: str
    :returns:
        tuple (username, password)
        WHERE
        username(str) is the user-supplied username.
        password(str) is the user-supplied password.
    :rtype: tuple
    """
    username = getpass.getpass(prompt=user_prompt)
    password = getpass.getpass(prompt=pwd_prompt)
    return (username, password)


def choose_from_list(options):
    """Iterate through a list of values and choose a value.

    :param options: List of options to choose from.
    :type options: list
    :returns:
        the user select option of None
    :rtype: option or None
    """
    exit = False
    result = False
    option = None
    x = 0
    print(f"These are the options: {', '.join(options)}.")
    while (exit == False and result is False and x < len(options)):
        option = options[x]
        prompt = f"Would you like to select '{option}'? "
        result = ask_yes_no(prompt=prompt, response_attempt=3)
        if (result is None or result is True):
            exit = True
        else:
            x += 1
    if result is True:
        return option
    else:
        return None


def truncate_value(value, length, suffix):
    """Truncate a string.

    :param value: String that should be truncated.
    :type value: str
    :param length: Final length of truncated string.
    :type length: int
    :param suffix: String that should be appended to truncated string.
    :type suffix: str
    :returns: the truncated string
    :rtype: str
    """
    if len(value) > length:
        if len(suffix) > length:
            suffix = suffix[:length]
        length2 = length - len(suffix)
        value = value[:length2] + suffix
    return value


# TODO this needs to be improved.
# TODO unittest.
def select_option(prompt, valid_response_set):
    """Select an option from a set of options.

    :param prompt: Message to display before displaying option.
    :type prompt: str
    :param valid_response_set: Set of valid options to choose.
    :type valid_response_set: set
    :returns: option
    :rtype: str, int
    """

    response_valid = False
    while response_valid == False:
        response = input(prompt)
        if response.isdigit():
            response = int(response)
        else:
            response = response.lower()

        if response in valid_response_set:
            response_valid = True
            if response == "y":
                response  = "yes"
            elif response == "n":
                response  = "no"
        else:
            print("Invalid response.")
    return response


# TODO unittest.
def get_values_from_dict_list(list_of_dicts):
    """Convert a list of dictionaries to a set of the dictionary values.

    :param list_of_dicts: List of dictionaries.
    :type list_of_dicts: list
    :returns: Set of values from all dictionaries in the list.
    :rtype: set
    """
    output_set = set()
    for dict in list_of_dicts:
        output_set = output_set | set(dict.values())
    return output_set


# TODO unittest.
def get_values_from_tuple_list(list_of_tuples):
    """Convert a list of tuples to a set of the tuple values.

    :param list_of_tuples: List of tuples.
    :type list_of_tuples: list
    :returns: Set of values from all tuples in the list.
    :rtype: set
    """
    output_set = set()
    for tup in list_of_tuples:
        output_set.add(tup[0])
    return output_set


def convert_list_to_dict(data_list, key):
    """Convert list of dictionaries to a dictionary of dictionaries

    :param data_list: List of dictionaries.
    :type data_list: list
    :param key: key in each dictionary to become the returned dictionary key.
    :type key: str
    :returns:
        Dictionary of all dictionaries. Returns an empty dictionary if
        all intended keys are not unique.
    :rtype: dict
    """
    data_dict = {}
    for element_dict in data_list:
        if element_dict[key] not in data_dict.keys():
            data_dict[element_dict[key]] = element_dict
    diff = len(data_list) - len(data_dict)
    if diff != 0:
        print("\nUnable to create dictionary since "
              "some intended keys are duplicated.")
        data_dict = {}
    return data_dict


# TODO unittest
def invert_dictionary(dictionary):
    """Inverts a dictionary, where the values and keys are swapped.

    :param dictionary: A dictionary to be inverted.
    :type dictionary: dict
    :returns: Returns an inverted dictionary of the given dictionary.
    :rtype: dict
    """
    new_dict = {}
    for key, value in dictionary.items():
        if isinstance(value, list):
            for subvalue in value:
                if new_dict.get(subvalue) is None:
                    new_dict[subvalue] = [key]
                else:
                    new_dict[subvalue].append(key)
        else:
            if new_dict.get(subvalue) is None:
                    new_dict[subvalue] = [key]
            else:
                new_dict[subvalue].append(key)

    return new_dict


# TODO unittest
def partition_list(data_list, size):
    """Chunks list into a list of lists with the given size.

    :param data_list: List to be split into equal-sized lists.
    :type data_list: list
    :param size: Length of the resulting list chunks.
    :param size: int
    :returns: Returns list of lists with length of the given size.
    :rtype: list[list]
    """
    chunked_list = [data_list[i*size:(i+1)*size]\
            for i in range((len(data_list) + size - 1) // size)]

    return chunked_list


# TODO unittest
def increment_histogram(data, histogram):
    """Increments a dictionary histogram based on given data.

    :param data: Data to be used to index or create new keys in the histogram.
    :type data: list
    :param histogram: Dictionary containing keys whose values contain counts.
    :type histogram: dict
    """
    for item in data:
        count = histogram.get(item, 0)
        count += 1
        histogram[item] = count


# TODO unittest
def sort_histogram_keys(histogram, descending=True):
    """Sorts a dictionary by its values and returns the sorted histogram.

    :param histogram: Dictionary containing keys whose values contain counts.
    :type histogram: dict
    :returns: A list from keys from the histogram sorted by value.
    :rtype: list
    """
    sorted_keys = [key for key, value in sorted(histogram.items(), 
                                                key=lambda item:item[1], 
                                                reverse=descending)] 

    return sorted_keys


def sort_histogram(histogram, descending=True):
    """Sorts a dictionary by its values and returns the sorted histogram.

    :param histogram: Dictionary containing keys whose values contain counts.
    :type histogram: dict
    :returns: An ordered dict from items from the histogram sorted by value.
    :rtype: OrderedDict
    """
    sorted_histogram = OrderedDict(sorted(histogram.items(),
                                          key=lambda item:item[1],
                                          reverse=descending))
    return sorted_histogram


# TODO unittest
def prepare_filepath(folder_path, file_name, folder_name=None):
    """Prepare path to new file.

    :param folder_path: Path to the directory to contain the file.
    :type folder_path: Path
    :param file_name: Name of the file.
    :type file_name: str
    :param folder_name: Name of sub-directory to create.
    :type folder_name: Path
    :returns: Path to file in directory.
    :rtype: Path
    """
    if folder_name is not None:
        subfolder_path = Path(folder_path, folder_name)
        subfolder_path.mkdir()
    else:
        subfolder_path = folder_path
    file_path = Path(subfolder_path, file_name)
    return file_path


def join_strings(input_list, delimiter=" "):
    """Open file and retrieve a dictionary of data.

    :param input_list: List of values to join.
    :type input_list: list
    :param delimiter: Delimiter used between values.
    :type delimiter: str
    :returns: Concatenated values, excluding all None and '' values.
    :rtype: str
    """
    concat_list = []
    for value in input_list:
        if (value is not None and value != ""):
            concat_list.append(value)
    if len(concat_list) > 0:
        string = delimiter.join(concat_list)
    else:
        string = ""
    return string


def merge_set_dicts(dict1, dict2):
    """Merge two dictionaries of sets.

    :param dict1: First dictionary of sets.
    :type dict1: dict
    :param dict2: Second dictionary of sets.
    :type dict2: dict
    :returns:
        Merged dictionary containing all keys from both dictionaries,
        and for each shared key the value is a set of merged values.
    :rtype: dict
    """
    keys = dict1.keys() | dict2.keys()
    dict3 = {}
    for key in keys:
        if key in dict1.keys():
            set1 = dict1[key]
        else:
            set1 = set()

        if key in dict2.keys():
            set2 = dict2[key]
        else:
            set2 = set()

        dict3[key] = set1 | set2
    return dict3


def choose_most_common(string, values):
    """Identify most common occurrence of several values in a string.

    :param string: String to search.
    :type string: str
    :param values:
        List of string characters. The order in the list indicates preference,
        in the case of a tie.
    :type values: list
    :returns: Value from values that occurs most.
    :rtype: str
    """
    # In case no value is found, default is to choose the first value.
    value1 = values[0]
    count1 = 0
    # Since order reflects ranking, with values deeper in the list
    # reflecting lower ranked values, reverse the order to search.
    for i in reversed(range(len(values))):
        value2 = values[i]
        count2 = string.count(value2)
        if count2 >= count1:
            value1 = value2
            count1 = count2
    return value1


def convert_to_encoded(values):
    """Converts a list of strings to utf-8 encoded values.

    :param values: Strings for a MySQL query to be encoded.
    :type values: list[str]
    :returns: List of utf-8 encoded values.
    :rtype: list[bytes]
    """
    if not isinstance(values, list):
        values = [values]

    converted_values = []
    for value in values:
        if value is None:
            converted_values.append(value)
        else:
            converted_values.append(value.encode("utf-8"))

    return converted_values


def convert_to_decoded(values):
    """Converts a list of strings to utf-8 encoded values.

    :param values: Byte values from MySQL queries to be decoded.
    :type values: list[bytes]
    :returns: List of utf-8 decoded values.
    :rtype: list[str]
    """
    if not isinstance(values, list):
        values = [values]

    converted_values = []
    for value in values:
        if value is None:
            converted_values.append(value)
        else:
            converted_values.append(value.decode("utf-8"))

    return converted_values


# TODO nearly identical to make_new_dir, so could probably be consolidated.
# TODO test
def make_new_file(output_dir, new_file, ext, attempt=1):
    """Make a new file.

    Checks to verify the new file name is valid and does not
    already exist. If it already exists, it attempts to extend
    the name with an integer suffix.

    :param output_dir:
        Full path to the directory where the new directory will be created.
    :type output_dir: Path
    :param new_file: Name of the new file to be created.
    :type new_file: Path
    :param ext: Name of the file extension to be used.
    :type ext: str
    :param attempt: Number of attempts to create the file.
    :type attempt: int
    :returns:
        If successful, the full path of the created file.
        If unsuccessful, None.
    :rtype: Path, None
    """
    valid = False
    count = 0
    while (not valid and count < attempt):
        if count > 0:
            new_file_mod = new_file.stem + "_" + str(count)
            new_file_mod = Path(new_file_mod)
        else:
            new_file_mod = new_file

        new_file_mod = new_file_mod + "." + ext
        new_path = Path(output_dir, new_file_mod)
        if not new_path.is_file():
            valid = True
        count += 1
    if not valid:
        return None
    else:
        return new_path


def clear_screen():
    """Brings the command line to the top of the screen."""

    if os.name == "nt":
        os.system("cls")
    else:
        os.system("clear")
